# Copyright (c) 2024 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
Modules
-------

The ``drgn.helpers.linux.module`` module contains helpers for working with
loaded kernel modules.
"""
import operator
from typing import Iterable, List, Tuple, Union

from drgn import NULL, IntegerLike, Object, ObjectNotFoundError, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.rbtree import rb_find

__all__ = (
    "address_to_module",
    "find_module",
    "for_each_module",
    "module_address_regions",
    "module_percpu_region",
    "module_taints",
)


@takes_program_or_default
def for_each_module(prog: Program) -> Iterable[Object]:
    """
    Returns all loaded kernel modules

    :returns: Iterable of ``struct module *`` objects
    """
    return list_for_each_entry("struct module", prog["modules"].address_of_(), "list")


@takes_program_or_default
def find_module(prog: Program, name: Union[str, bytes]) -> Object:
    """
    Lookup a kernel module by name, or return NULL if not found

    :param name: name to search for
    :returns: the ``struct module *`` by that name, or NULL
    """
    if isinstance(name, str):
        name = name.encode()
    for module in for_each_module(prog):
        if module.name.string_() == name:
            return module
    return NULL(prog, "struct module *")


def module_percpu_region(mod: Object) -> Tuple[int, int]:
    """
    Lookup the percpu memory region of a module.

    Given a ``struct module *``, return the address (as a an int) and the length
    of the percpu memory region. Modules may have a NULL percpu region, in which
    case (0, 0) is returned. Rarely, on kernels without ``CONFIG_SMP``, there is
    no percpu region at all, and this function returns (0, 0)

    :param mod: Object of type ``struct module *``
    :returns: (base, size) of the module percpu region
    """
    try:
        return mod.percpu.value_(), mod.percpu_size.value_()
    except AttributeError:
        return 0, 0


def _range_from_module_layout(layout: Object) -> Tuple[int, int]:
    # For "struct module_layout" (old) or "struct module_memory"
    return layout.base.value_(), layout.size.value_()


def _range_from_module(module: Object, kind: str) -> Tuple[int, int]:
    # For reading a range directly from "struct module" (old kernels)
    return (
        module.member_(f"module_{kind}").value_(),
        module.member_(f"{kind}_size").value_(),
    )


def _ranges_from_module_memory(mod: Object) -> List[Tuple[int, int]]:
    # For reading all ranges from a modules "struct module_memory"
    return [_range_from_module_layout(mem) for mem in mod.mem]


def module_address_regions(mod: Object) -> List[Tuple[int, int]]:
    """
    Returns a list of address ranges for a module

    Given a ``struct module *``, return every address range associated with the
    module. Note that the number of address ranges and their interpretations
    vary across kernel versions. Some kernel versions provide additional
    information about some regions (e.g. text, data, R/O, init). This API
    doesn't distinguish. However, this API does not provide the module's percpu
    region: use ``module_percpu_region()`` for that.

    :param mod: Object of type ``struct module *``
    :returns: list of tuples: (starting memory address, length of address range)
    """
    try:
        # Since Linux 6.4, ac3b432839234 ("module: replace module_layout with
        # module_memory"), module address regions are broken into several types,
        # each with their own base and size.
        mod.prog_.constant("MOD_MEM_NUM_TYPES")
    except LookupError:
        pass
    else:
        return _ranges_from_module_memory(mod)

    try:
        # Prior to 6.4, there were two "struct module_layout" objects,
        # core_layout and init_layout, which contained the module's memory
        # layout and any memory which could be freed after init. The init_layout
        # is usually NULL / size 0. The module_layout structure has more
        # information to say where text ends, where rodata ends, etc. We ignore
        # these.
        core = _range_from_module_layout(mod.core_layout)
        init = _range_from_module_layout(mod.init_layout)
    except AttributeError:
        # Prior to 4.5, 7523e4dc5057 ("module: use a structure to encapsulate
        # layout."), the layout information was stored as variables directly in
        # the struct module. They were prefixed with "core_" and "init_".
        core = _range_from_module(mod, "core")
        init = _range_from_module(mod, "init")

    ret = [core]
    if init:
        ret.append(init)
    return ret


def _addrmod_tree(mod_tree: Object, addr: int) -> Object:
    prog = mod_tree.prog_

    # The module tree is "latched": there are two parallel trees. Which one is
    # in use depends on the seqcount, which gets incremented for each
    # modification. This is a really neat approach that allows reads in parallel
    # with a writer. In our use case, it's probably not worth verifying the
    # seqcount after the fact. What we do need is the index (0 or 1). This may
    # be a seqcount_latch_t, or before 24bf401cebfd6 ("rbtree_latch: Use
    # seqcount_latch_t"), a regular seqcount_t.
    try:
        idx = mod_tree.root.seq.seqcount.sequence.value_() & 1
    except AttributeError:
        idx = mod_tree.root.seq.sequence.value_() & 1

    # In ac3b432839234 ("module: replace module_layout with module_memory"),
    # struct module_layout was replaced by module_memory. The module_layout
    # encoded the separate regions (text, data, rodata, etc) in a single
    # structure, whereas module_memory is a simple base pointer followed by a
    # size: one module_memory structure is used per kind of memory. However,
    # both of them contain a "base" pointer that indicates the start of the
    # region, a "size" that indicates its total size, and a "mtn.mod" pointer
    # which refers to the relevant module. So for our use case, they are
    # interchangeable, except for their names.
    try:
        tp = prog.type("struct module_memory")
    except LookupError:
        tp = prog.type("struct module_layout")

    def cmp(v: int, node: Object) -> int:
        start = node.base.value_()
        end = start + node.size.value_()
        if v < start:
            return -1
        elif v >= end:
            return 1
        else:
            return 0

    mem = rb_find(
        tp,
        mod_tree.root.tree[idx].address_of_(),
        f"mtn.node.node[{idx}]",  # container_of allows array indices!
        addr,
        cmp,
    )
    if mem:
        return mem.mtn.mod
    else:
        return NULL(prog, "struct module *")


@takes_program_or_default
def address_to_module(prog: Program, addr: IntegerLike) -> Object:
    """
    Return the ``struct module *`` associated with a memory address

    If the address is a text, data, or read-only data address associated with a
    kernel module, then this function returns the module it is associated with.
    Otherwise, returns NULL. Note that dynamic memory (e.g. slab objects)
    generally can't be associated with the module that allocated it. Further,
    static & dynamic per-cpu address cannot be associated with their associated
    module either.

    Normally, this lookup is efficient, thanks to
    ``CONFIG_MODULES_TREE_LOOKUP``, which provides a red-black tree of module
    address ranges, and is `very commonly`__ enabled. However, on some uncommon
    configurations the rbtree may not be present. In those cases, we fall back
    to a linear search of each kernel module's memory regions.

    .. __: https://oracle.github.io/kconfigs/?config=MODULES_TREE_LOOKUP&config=UTS_RELEASE

    :param addr: memory address to lookup
    :returns: the ``struct module *`` associated with the memory, or NULL
    """
    addr = operator.index(addr)
    try:
        mod_tree = prog["mod_tree"]
    except LookupError:
        pass
    else:
        return _addrmod_tree(mod_tree, addr)

    for module in for_each_module(prog):
        for start, length in module_address_regions(module):
            if start <= addr < start + length:
                return module

    return NULL(prog, "struct module *")


def module_taints(module: Object) -> str:
    """
    Get a kernel module's taint flags as a string.

    If the module did not taint the kernel, then this returns an empty string:

    >>> module_taints(module)
    ''

    Otherwise, it returns the flags as letters (without spaces):

    >>> module_taints(module)
    'O'

    See the `kernel documentation
    <https://docs.kernel.org/admin-guide/tainted-kernels.html>`_ for an
    explanation of the flags.

    :param module: ``struct module *``
    """
    mask = module.taints.value_()
    if not mask:
        return ""

    # Before Linux kernel commit 37ade54f386c ("taint/module: remove
    # unnecessary taint_flag.module field") (in v6.19), there was a distinction
    # between what taint flags could and couldn't be set on a module. The
    # distinction is unnecessary for us since we don't need a statically-sized
    # buffer.
    parts = []
    try:
        taint_flags = module.prog_["taint_flags"]
    except ObjectNotFoundError:
        # Before Linux kernel commit 7fd8329ba502 ("taint/module: Clean up
        # global and module taint flags handling") (in v4.10), the array and
        # members had different names and the bit number was a member.
        for t in module.prog_["tnts"]:
            if mask & (1 << t.bit.value_()):
                parts.append(chr(t.true))  # type: ignore[arg-type]  # python/typeshed#13494
    else:
        for i, t in enumerate(taint_flags):
            if mask & (1 << i):
                parts.append(chr(t.c_true))  # type: ignore[arg-type]  # python/typeshed#13494
    return "".join(parts)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash ipcs command."""

import argparse
from typing import Any, Callable, Iterable, Iterator, List, Sequence, Tuple

from drgn import Object, Program
from drgn.commands import argument, argument_group, drgn_argument
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    crash_command,
    crash_get_context,
)
from drgn.helpers.common.format import CellFormat, print_table
from drgn.helpers.linux.ipc import (
    decode_sysv_shm_mode_flags,
    find_sysv_msg_queue,
    find_sysv_sem_array,
    find_sysv_shm,
    for_each_sysv_msg_queue,
    for_each_sysv_sem_array,
    for_each_sysv_shm,
)
from drgn.helpers.linux.user import kuid_val


def _append_shm_code(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.user", "kuid_val")
    code.add_from_import("drgn.helpers.linux.ipc", "decode_sysv_shm_flags")
    code.append(
        """\
    key = shm.shm_perm.key
    id = shm.shm_perm.id
    uid = kuid_val(shm.shm_perm.uid)
    perms = shm.shm_perm.mode & 0o777
    bytes = shm.shm_segsz
    nattch = shm.shm_nattch
    status = decode_sysv_shm_flags(shm)
"""
    )


def _print_shms(shms: Iterable[Object]) -> None:
    rows: List[Sequence[Any]] = [
        (
            "SHMID_KERNEL",
            "KEY",
            "SHMID",
            "UID",
            "PERMS",
            "BYTES",
            "NATTCH",
            "STATUS",
        )
    ]
    for shm in shms:
        mode: int = shm.shm_perm.mode.value_()
        status = decode_sysv_shm_mode_flags(shm.prog_, mode)
        if status == "0":
            status = ""
        rows.append(
            (
                CellFormat(shm.value_(), "<x"),
                f"{shm.shm_perm.key.value_():08x}",
                CellFormat(shm.shm_perm.id.value_(), "<"),
                CellFormat(kuid_val(shm.shm_perm.uid), "<"),
                CellFormat(mode & 0o777, "<o"),
                CellFormat(shm.shm_segsz.value_(), "<"),
                CellFormat(shm.shm_nattch.value_(), "<"),
                status,
            )
        )
    print_table(rows)
    if len(rows) == 1:
        print("(none allocated)")


def _append_sem_array_code(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.user", "kuid_val")
    code.append(
        """\
    key = sem_array.sem_perm.key
    id = sem_array.sem_perm.id
    uid = kuid_val(sem_array.sem_perm.uid)
    perms = sem_array.sem_perm.mode
    nsems = sem_array.sem_nsems
"""
    )


def _print_sem_arrays(sem_arrays: Iterable[Object]) -> None:
    rows: List[Sequence[Any]] = [
        (
            "SEM_ARRAY",
            "KEY",
            "SEMID",
            "UID",
            "PERMS",
            "NSEMS",
        )
    ]
    for sem_array in sem_arrays:
        rows.append(
            (
                CellFormat(sem_array.value_(), "<x"),
                f"{sem_array.sem_perm.key.value_():08x}",
                CellFormat(sem_array.sem_perm.id.value_(), "<"),
                CellFormat(kuid_val(sem_array.sem_perm.uid), "<"),
                CellFormat(sem_array.sem_perm.mode.value_() & 0o777, "<o"),
                CellFormat(sem_array.sem_nsems.value_(), "<"),
            )
        )
    print_table(rows)
    if len(rows) == 1:
        print("(none allocated)")


def _append_msg_queue_code(code: CrashDrgnCodeBuilder) -> None:
    code.add_from_import("drgn.helpers.linux.user", "kuid_val")
    code.append(
        """\
    key = msg_queue.q_perm.key
    id = msg_queue.q_perm.id
    uid = kuid_val(msg_queue.q_perm.uid)
    perms = msg_queue.q_perm.mode
    used_bytes = msg_queue.q_cbytes
    messages = msg_queue.q_qnum
"""
    )


def _print_msg_queues(msg_queues: Iterable[Object]) -> None:
    rows: List[Sequence[Any]] = [
        (
            "MSG_QUEUE",
            "KEY",
            "MSQID",
            "UID",
            "PERMS",
            "USED-BYTES",
            "MESSAGES",
        )
    ]
    for msg_queue in msg_queues:
        rows.append(
            (
                CellFormat(msg_queue.value_(), "<x"),
                f"{msg_queue.q_perm.key.value_():08x}",
                CellFormat(msg_queue.q_perm.id.value_(), "<"),
                CellFormat(kuid_val(msg_queue.q_perm.uid), "<"),
                CellFormat(msg_queue.q_perm.mode.value_() & 0o777, "<o"),
                CellFormat(msg_queue.q_cbytes.value_(), "<"),
                CellFormat(msg_queue.q_qnum.value_(), "<"),
            )
        )
    print_table(rows)
    if len(rows) == 1:
        print("(none allocated)")


@crash_command(
    description="System V IPC",
    arguments=(
        argument_group(
            argument(
                "-s",
                dest="sem_arrays",
                action="store_true",
                help="display semaphore arrays",
            ),
            argument(
                "-m",
                dest="shms",
                action="store_true",
                help="display shared memory segments",
            ),
            argument(
                "-q",
                dest="msg_queues",
                action="store_true",
                help="display message queues",
            ),
            title="types",
            description="IPC types to display. If none are given, all three are displayed.",
        ),
        argument(
            "-n",
            dest="task",
            metavar="pid|task",
            type="pid_or_task",
            help="""
            search in the IPC namespace of a task, given as either a decimal
            process ID or a hexadecimal ``task_struct`` address. Defaults to
            the IPC namespace of the current context
            """,
        ),
        argument(
            "id_or_addr",
            metavar="id|addr",
            nargs="*",
            help="""
            search for IPC instances with the given decimal identifier or
            hexadecimal kernel address. May be given multiple times
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_ipcs(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    if not any((args.sem_arrays, args.shms, args.msg_queues)):
        args.sem_arrays = args.shms = args.msg_queues = True

    types: List[
        Tuple[
            str,
            Callable[[CrashDrgnCodeBuilder], None],
            Callable[[Iterable[Object]], None],
            Callable[[Object], Iterator[Object]],
            Callable[[Object, int], Object],
        ]
    ] = []
    if args.shms:
        types.append(
            ("shm", _append_shm_code, _print_shms, for_each_sysv_shm, find_sysv_shm)
        )
    if args.sem_arrays:
        types.append(
            (
                "sem_array",
                _append_sem_array_code,
                _print_sem_arrays,
                for_each_sysv_sem_array,
                find_sysv_sem_array,
            )
        )
    if args.msg_queues:
        types.append(
            (
                "msg_queue",
                _append_msg_queue_code,
                _print_msg_queues,
                for_each_sysv_msg_queue,
                find_sysv_msg_queue,
            )
        )

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)

        # Avoid the context/IPC namespace noise if -n wasn't given and the
        # current context is in the initial IPC namespace.
        if (
            args.task is None
            and crash_get_context(prog).nsproxy.ipc_ns
            == prog["init_ipc_ns"].address_of_()
        ):
            ns_var = ""
            ns_var_with_comma = ""
        else:
            code.append_crash_context(args.task)
            code.append("ipc_ns = task.nsproxy.ipc_ns\n\n")
            ns_var = "ipc_ns"
            ns_var_with_comma = "ipc_ns, "

        first = True
        if args.id_or_addr:
            for id_or_addr in args.id_or_addr:
                try:
                    id = int(id_or_addr, 10)
                except ValueError:
                    addr = int(id_or_addr, 16)
                    code.append(f"addr = {hex(addr)}\n\n")
                    for name, append_fn, _, _, _ in types:
                        if first:
                            first = False
                        else:
                            code.append("\n")
                        code.add_from_import(
                            "drgn.helpers.linux.ipc", "for_each_sysv_" + name
                        )
                        code.append(
                            f"""\
for {name} in for_each_sysv_{name}({ns_var}):
    if {name}.value_() != addr:
        continue
"""
                        )
                        append_fn(code)
                else:
                    for name, append_fn, _, _, _ in types:
                        if first:
                            first = False
                        else:
                            code.append("\n")
                        code.add_from_import(
                            "drgn.helpers.linux.ipc", "find_sysv_" + name
                        )
                        code.append(
                            f"""\
{name} = find_sysv_{name}({ns_var_with_comma}{id})
if {name}:
"""
                        )
                        append_fn(code)
        else:
            for name, append_fn, _, _, _ in types:
                if first:
                    first = False
                else:
                    code.append("\n")
                code.add_from_import("drgn.helpers.linux.ipc", "for_each_sysv_" + name)
                code.append(
                    f"""\
for {name} in for_each_sysv_{name}({ns_var}):
"""
                )
                append_fn(code)

        return code.print()

    ns = crash_get_context(prog, args.task).nsproxy.ipc_ns

    first = True
    if args.id_or_addr:
        for id_or_addr in args.id_or_addr:
            arg_found = False
            try:
                id = int(id_or_addr, 10)
            except ValueError:
                addr = int(id_or_addr, 16)
                for _, _, print_fn, for_each_fn, _ in types:
                    matches = [obj for obj in for_each_fn(ns) if obj.value_() == addr]
                    if matches:
                        arg_found = True
                        if first:
                            first = False
                        else:
                            print()
                        print_fn(matches)
                if not arg_found:
                    if first:
                        first = False
                    else:
                        print()
                    print("invalid address:", hex(addr))
            else:
                for _, _, print_fn, _, find_fn in types:
                    found = find_fn(ns, id)
                    if found:
                        arg_found = True
                        if first:
                            first = False
                        else:
                            print()
                        print_fn((found,))
                if not arg_found:
                    if first:
                        first = False
                    else:
                        print()
                    print("invalid id:", id)
    else:
        for _, _, print_fn, for_each_fn, _ in types:
            if first:
                first = False
            else:
                print()
            print_fn(for_each_fn(ns))



#!/usr/bin/env drgn
# Copyright (c) 2025, Kylin Software, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import sys
import argparse
from drgn.helpers.linux import find_task
from drgn import stack_trace

def main():
    parser = argparse.ArgumentParser(description="Show stack trace and optionally local variables of a specific frame")
    parser.add_argument("pid", type=int, help="Target process PID")
    parser.add_argument("frameid", type=int, nargs="?", help="Stack frame index (optional)")
    args = parser.parse_args()

    task = find_task(args.pid)
    print(f"PID {args.pid} process name: {task.comm.string_().decode(errors='replace')}")
    trace = stack_trace(task)
    print(trace)
    if args.frameid is not None:
        frame = trace[int(args.frameid)]
        print(frame)
        locals_ = frame.locals()
        print(locals_)
        for var in locals_:
            try:
                value = frame[var]
                print(f"{var}: {value}")
            except KeyError:
                print(f"{var}: Not found")

if __name__ == "__main__":
    main()



#!/usr/bin/env drgn
# Copyright (c) 2024, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Print key details of the filesystems in the btrfs fs_uuids cache

Usage:
* drgn -s <btrfs.ko.debug> -s <vmlinux> -c <vmcore> btrfs_print_fs_uuids_cache.py
    - This only prints key details of the filesystems
* drgn -s <btrfs.ko.debug> -s <vmlinux> -c <vmcore> btrfs_print_fs_uuids_cache.py -d yes
    - This dumps all the structures. The output will be enormous,
    so it's recommended to redirect it to a file and read it from there.
"""

import drgn
import argparse

from drgn.helpers.common import *
from drgn.helpers.linux import (
    list_for_each_entry,
    for_each_possible_cpu,
    per_cpu_ptr,
)

parser = argparse.ArgumentParser()
parser.add_argument("-d", "--dump", required = False, help = "Dump all structure data")

args = parser.parse_args()

def format(uuid_list):
    """Helper to format uuid from list representation"""
    part = ""
    for ele in uuid_list:
        part += "%02x" %ele
    return part

def get_uuid_from_list(uuid_list):
    """Helper to format uuid from list representation"""
    uuid = ""
    uuid += format(uuid_list[0:4]) + "-"
    uuid += format(uuid_list[4:6]) + "-"
    uuid += format(uuid_list[6:8]) + "-"
    uuid += format(uuid_list[8:10]) + "-"
    uuid += format(uuid_list[10:])
    return uuid

def print_mount_info(mnt):
    """Print key fields of `struct mount`"""
    print(f"Mount Info:")
    print(f"\tMount Point: {mnt.mnt_mountpoint.d_iname.string_().decode('utf-8')}")
    print(f"\tVFS Mount Flags: {mnt.mnt.mnt_flags}")
    print(f"\tDevice Name: {mnt.mnt_devname.string_().decode('utf-8')}")
    if mnt.mnt_mp:
        print(f"\tMount Point Count: {mnt.mnt_mp.m_count.value_()}")
    else:
        print("\tMount Point is NULL")
    print(f"\tMount ID: {mnt.mnt_id}")
    print(f"\tMount Group ID: {mnt.mnt_group_id}")
    print(f"\tMount Expiry Mark: {mnt.mnt_expiry_mark}")
    return

def print_super_block_info(sb):
    """Print key fields of `struct super_block`"""
    print("Super Block:")
    if not sb:
        print("\tSuper block is NULL")
        return
    print(f"\tsb ref count: {sb.s_active.counter}")
    print(f"\tsb s_count: {sb.s_count}")
    print(f"\tsb umount rw_sem counter {sb.s_umount.count.counter}")
    print(f"\tsb s_flags: {sb.s_flags}")
    print(f"\tsb s_dev {sb.s_dev}")
    print(f"\tsb s_id {sb.s_id}")
    print(f"\tsb i_flags {sb.s_iflags}")

    for mnt in list_for_each_entry("struct mount", sb.s_mounts.address_of_(), "mnt_instance"):
        if not mnt.mnt_master:
            print_mount_info(mnt)
            break

    return

def print_fs_info(fs_info):
    """Print key fields of `struct btrfs_fs_info`"""
    print("FS Info:")
    if not fs_info:
        print("\tfs_info is NULL")
        return

    print(f"\tFS State: {fs_info.fs_state}")
    print(f"\tFlags: {fs_info.flags}")
    print(f"\tmount opt: {fs_info.mount_opt}")

    print_super_block_info(fs_info.sb)
    return

def print_fs_devices_info(fs_dev):
    """"Print key fields of `struct btrfs_fs_devices`"""
    print("FS Devices:")
    print(f"\tFS UUID: {get_uuid_from_list(list(fs_dev.fsid))}")
    print(f"\tMETADATA UUID: {get_uuid_from_list(list(fs_dev.metadata_uuid))}")
    print(f"\tnum_devices: {fs_dev.num_devices}")
    print(f"\topen_devices: {fs_dev.open_devices}")
    print(f"\trw_devices: {fs_dev.rw_devices}")
    print(f"\tmissing_devices: {fs_dev.missing_devices}")
    print(f"\ttotal_rw_bytes: {fs_dev.total_rw_bytes}")
    print(f"\ttotal_devices: {fs_dev.total_devices}")
    print(f"\tlatest_generation: {fs_dev.latest_generation}")
    print(f"\topened: {fs_dev.opened}")
    try:
        print(f"\tfsid_change: {fs_dev.fsid_change}")
    except AttributeError:
        pass
    return

def print_dev_info(dev):
    """Print key fields of `struct btrfs_device`"""
    print("Device Info:")
    print(f"\tName: {str(dev.name.str)}")
    print(f"\tstate: {dev.dev_state}")
    print(f"\tdevid: {dev.devid}")
    print(f"\tgeneration: {dev.generation}")
    print(f"\ttype: {dev.type}")
    if not dev.bdev:
        print(f"\tbdev: NULL")
    else:
        print(f"\tbd_dev: {dev.bdev.bd_dev}")
        print(f"\tbd_partno: {dev.bdev.bd_partno}")
    return

def dump_mounts(s_mounts):
    """Dump master `struct mount` of a mount point"""
    for mnt in list_for_each_entry("struct mount", s_mounts.address_of_(), "mnt_instance"):
        if not mnt.mnt_master:
            print(f"struct mount:\n{mnt}")
            print(f"Mount Point Dentry:\n{mnt.mnt_mountpoint}")
            print(f"Namespace:\n{mnt.mnt_ns}")
            print(f"Mount Point Struct:\n{mnt.mnt_mp}")
            for cpu in for_each_possible_cpu(prog):
                print(f"mnt_pcp {cpu}:\n{per_cpu_ptr(mnt.mnt_pcp, cpu)}")
            break
    return

def dump_all(fs_uuids):
    """
    Dump all structure data of:
        - struct btrfs_fs_devices
        - struct btrfs_device
        - struct btrfs_fs_info
        - Running transaction => struct btrfs_transaction
        - struct btrfs_super_block
        - struct super_block
        - struct file_system_type
        - Root Dentry => struct dentry
        - Root Inode => struct inode
        - struct user_namespace
    """
    for fs_dev in list_for_each_entry("struct btrfs_fs_devices", fs_uuids.address_of_(), "fs_list"):
        print("-"*30 + f"{get_uuid_from_list(list(fs_dev.fsid))}" + '-'*30)
        print("FS Devices:\n", fs_dev)
        print("fs_dev.latest_bdev:", fs_dev.latest_bdev)
        if fs_dev.fs_info:
            print("FS Info:\n", fs_dev.fs_info)
            print("fs_info Running transaction:\n", fs_dev.fs_info.running_transaction)
            print("Disk superblock:\n", fs_dev.fs_info.super_copy)
            if fs_dev.fs_info.sb:
                print("Superblock:\n", fs_dev.fs_info.sb)
                print("Fs Type:\n", fs_dev.fs_info.sb.s_type)
                print("Root Dentry:\n", fs_dev.fs_info.sb.s_root)
                print("Root Inode:\n", fs_dev.fs_info.sb.s_root.d_inode)
                print("sb user namespace:\n", fs_dev.fs_info.sb.s_user_ns)
                dump_mounts(fs_dev.fs_info.sb.s_mounts)
            else:
                print("Superblock: NULL")
        else:
            print("FS Info: NULL")

        for dev in list_for_each_entry("struct btrfs_device", fs_dev.devices.address_of_(), "dev_list"):
            print("Device:\n", dev)
            print("Block Device:\n", dev.bdev)
            print("dev.flush_bio:\n", dev.flush_bio) 
        print("\n")
    return

def print_btrfs_cache(fs_uuids):
    """
    Iterate through the fs_uuids cache and print the details of each filesystem.
    """
    for fs_dev in list_for_each_entry("struct btrfs_fs_devices", fs_uuids.address_of_(), "fs_list"):
        print("-"*60)
        print_fs_devices_info(fs_dev)
        print_fs_info(fs_dev.fs_info)

        for dev in list_for_each_entry("struct btrfs_device", fs_dev.devices.address_of_(), "dev_list"):
            print_dev_info(dev)
    return

fs_uuids = prog['fs_uuids']

if args.dump == "yes":
    dump_all(fs_uuids)
else:
    print_btrfs_cache(fs_uuids)


#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Search kernel memory for a byte string. This only works if the kernel was built
with CONFIG_PROC_KCORE=y.
"""

import argparse
import sys

from drgn import FaultError, Object
from drgn.helpers.common.memory import identify_address
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.mm import for_each_vmap_area, virt_to_page


def virt_to_vmap_address(prog, addr):
    page = virt_to_page(addr)
    for va in for_each_vmap_area(prog):
        vm = va.vm.read_()
        if vm:
            for i, va_page in enumerate(
                Object(
                    prog, prog.array_type(page.type_, vm.nr_pages), address=vm.pages
                ).read_()
            ):
                if va_page == page:
                    return (
                        va.va_start.value_()
                        + (i << prog["PAGE_SHIFT"])
                        + (addr & (prog["PAGE_SIZE"].value_() - 1))
                    )
    return None


def search_memory(prog, needle):
    if isinstance(needle, Object):
        needle = needle.to_bytes_()

    KCORE_RAM = prog["KCORE_RAM"]
    CHUNK_SIZE = 1024 * 1024
    PAGE_SIZE = prog["PAGE_SIZE"].value_()
    for kc in list_for_each_entry(
        "struct kcore_list", prog["kclist_head"].address_of_(), "list"
    ):
        if kc.type != KCORE_RAM:
            continue
        addr = kc.addr.value_()
        end = addr + kc.size.value_()
        while addr < end:
            try:
                buf = prog.read(addr, min(CHUNK_SIZE, end - addr))
            except FaultError:
                # We start with a large chunk size to reduce the overhead of
                # reading memory. However, if we're reading from a core dump,
                # reading with a large chunk size may fault on excluded pages.
                if CHUNK_SIZE > PAGE_SIZE:
                    # We faulted with a large chunk size. Fall back to
                    # page-by-page and retry.
                    CHUNK_SIZE = PAGE_SIZE
                else:
                    # We're already reading page-by-page. Skip this page.
                    addr += CHUNK_SIZE
                continue

            i = 0
            while i < len(buf):
                i = buf.find(needle, i)
                if i < 0:
                    break

                vmap_address = virt_to_vmap_address(prog, addr + i)
                if vmap_address is not None:
                    identity = identify_address(prog, vmap_address)
                else:
                    identity = identify_address(prog, addr + i)

                if identity is None:
                    print(hex(addr + i))
                else:
                    print(hex(addr + i), identity)
                i += 1

            addr += CHUNK_SIZE


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Search kernel memory for a value and print the addresses where it is found"
    )
    group = parser.add_argument_group(
        title="search target",
        description="If none of these are given, search for a byte string read from standard input.",
    ).add_mutually_exclusive_group()
    group.add_argument("--string", help="search for an ASCII/UTF-8 string")
    group.add_argument("--hex", help="search for a byte string, given in hexadecimal")
    group.add_argument(
        "--address", help="search for an address-sized integer, given in hexadecimal"
    )
    args = parser.parse_args()

    if args.string is not None:
        needle = args.string.encode()
    elif args.hex is not None:
        needle = bytes.fromhex(args.hex)
    elif args.address is not None:
        needle = Object(prog, "void *", int(args.address, 16))
    else:
        needle = sys.stdin.buffer.read()

    search_memory(prog, needle)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later


"""
HugeTLB
-------

The ``drgn.helpers.linux.hugetlb`` module provides helpers for working with
HugeTLB pages.
"""

from typing import Iterator

from drgn import Object, ObjectNotFoundError, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.mm import PageUsage

__all__ = (
    "for_each_hstate",
    "huge_page_size",
    "hugetlb_total_pages",
    "hugetlb_total_usage",
)


@takes_program_or_default
def for_each_hstate(prog: Program) -> Iterator[Object]:
    """
    Iterate over all HugeTLB page size states.

    >>> [h.name.string_() for h in for_each_hstate()]
    [b'hugepages-1048576kB', b'hugepages-2048kB']

    :return: Iterator of ``struct hstate *`` objects.
    """
    try:
        hstates = prog["hstates"]
    except ObjectNotFoundError:
        # CONFIG_HUGETLBFS=n.
        return iter(())
    return (h.address_of_() for h in hstates[: prog["hugetlb_max_hstate"]])


def huge_page_size(hstate: Object) -> Object:
    """
    Return the size of a HugeTLB state in bytes.

    :param hstate: ``struct hstate *``
    :return: ``unsigned long``
    """
    return hstate.prog_["PAGE_SIZE"] << hstate.order


@takes_program_or_default
def hugetlb_total_pages(prog: Program) -> int:
    """Get the total number of HugeTLB pages (in ``PAGE_SIZE`` units)."""
    return sum(
        h.nr_huge_pages.value_() << h.order.value_() for h in for_each_hstate(prog)
    )


@takes_program_or_default
def hugetlb_total_usage(prog: Program) -> PageUsage:
    """
    Get the total number of HugeTLB pages and the number of free HugeTLB pages
    (in ``PAGE_SIZE`` units).
    """
    pages = 0
    free_pages = 0
    for h in for_each_hstate(prog):
        order = h.order.value_()
        pages += h.nr_huge_pages.value_() << order
        free_pages += h.free_huge_pages.value_() << order
    return PageUsage(pages=pages, free_pages=free_pages)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Red-Black Trees
---------------

The ``drgn.helpers.linux.rbtree`` module provides helpers for working with
red-black trees from :linux:`include/linux/rbtree.h`.
"""

from typing import Callable, Generator, Iterator, Tuple, TypeVar, Union

from drgn import NULL, Object, Type, container_of
from drgn.helpers import ValidationError

__all__ = (
    "RB_EMPTY_ROOT",
    "RB_EMPTY_NODE",
    "rb_find",
    "rb_first",
    "rb_last",
    "rb_next",
    "rb_parent",
    "rb_prev",
    "rbtree_inorder_for_each",
    "rbtree_inorder_for_each_entry",
    "rbtree_preorder_for_each",
    "rbtree_preorder_for_each_entry",
    "validate_rbtree",
    "validate_rbtree_inorder_for_each_entry",
)


def RB_EMPTY_ROOT(root: Object) -> bool:
    """
    Return whether a red-black tree is empty.

    :param node: ``struct rb_root *``
    """
    return not root.rb_node


def RB_EMPTY_NODE(node: Object) -> bool:
    """
    Return whether a red-black tree node is empty, i.e., not inserted in a
    tree.

    :param node: ``struct rb_node *``
    """
    return node.__rb_parent_color.value_() == node.value_()


def rb_parent(node: Object) -> Object:
    """
    Return the parent node of a red-black tree node.

    :param node: ``struct rb_node *``
    :return: ``struct rb_node *``
    """
    return Object(node.prog_, node.type_, value=node.__rb_parent_color.value_() & ~3)


# Return parent node and whether the node is black.
def _rb_parent_color(node: Object) -> Tuple[Object, bool]:
    value = node.__rb_parent_color.value_()
    return Object(node.prog_, node.type_, value=value & ~3), (value & 1) != 0


def rb_first(root: Object) -> Object:
    """
    Return the first node (in sort order) in a red-black tree, or ``NULL`` if
    the tree is empty.

    :param root: ``struct rb_root *``
    :return: ``struct rb_node *``
    """
    node = root.rb_node.read_()
    if not node:
        return node
    while True:
        next = node.rb_left.read_()
        if not next:
            return node
        node = next


def rb_last(root: Object) -> Object:
    """
    Return the last node (in sort order) in a red-black tree, or ``NULL`` if
    the tree is empty.

    :param root: ``struct rb_root *``
    :return: ``struct rb_node *``
    """
    node = root.rb_node.read_()
    if not node:
        return node
    while True:
        next = node.rb_right.read_()
        if not next:
            return node
        node = next


def rb_next(node: Object) -> Object:
    """
    Return the next node (in sort order) after a red-black node, or ``NULL`` if
    the node is the last node in the tree or is empty.

    :param node: ``struct rb_node *``
    :return: ``struct rb_node *``
    """
    node = node.read_()

    if RB_EMPTY_NODE(node):
        return NULL(node.prog_, node.type_)

    next = node.rb_right.read_()
    if next:
        node = next
        while True:
            next = node.rb_left.read_()
            if not next:
                return node
            node = next

    parent = rb_parent(node).read_()
    while parent and node == parent.rb_right:
        node = parent
        parent = rb_parent(node).read_()
    return parent


def rb_prev(node: Object) -> Object:
    """
    Return the previous node (in sort order) before a red-black node, or
    ``NULL`` if the node is the first node in the tree or is empty.

    :param node: ``struct rb_node *``
    :return: ``struct rb_node *``
    """
    node = node.read_()

    if RB_EMPTY_NODE(node):
        return NULL(node.prog_, node.type_)

    next = node.rb_left.read_()
    if next:
        node = next
        while True:
            next = node.rb_right.read_()
            if not next:
                return node
            node = next

    parent = rb_parent(node).read_()
    while parent and node == parent.rb_left:
        node = parent
        parent = rb_parent(node).read_()
    return parent


def rbtree_inorder_for_each(root: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a red-black tree, in sort order.

    :param root: ``struct rb_root *``
    :return: Iterator of ``struct rb_node *`` objects.
    """

    def aux(node: Object) -> Iterator[Object]:
        if node:
            yield from aux(node.rb_left.read_())
            yield node
            yield from aux(node.rb_right.read_())

    yield from aux(root.rb_node.read_())


def rbtree_inorder_for_each_entry(
    type: Union[str, Type], root: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a red-black tree in sorted order.

    :param type: Entry type.
    :param root: ``struct rb_root *``
    :param member: Name of ``struct rb_node`` member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = root.prog_.type(type)
    for node in rbtree_inorder_for_each(root):
        yield container_of(node, type, member)


def rbtree_preorder_for_each(root: Object) -> Iterator[Object]:
    """
    Iterate over all of the nodes in a red-black tree, visiting each node
    before its children.

    :param root: ``struct rb_root *``
    :return: Iterator of ``struct rb_node *`` objects.
    """

    def aux(node: Object) -> Iterator[Object]:
        if node:
            yield node
            yield from aux(node.rb_left.read_())
            yield from aux(node.rb_right.read_())

    yield from aux(root.rb_node.read_())


def rbtree_preorder_for_each_entry(
    type: Union[str, Type], root: Object, member: str
) -> Iterator[Object]:
    """
    Iterate over all of the entries in a red-black tree, visiting each node
    before its children.

    :param type: Entry type.
    :param root: ``struct rb_root *``
    :param member: Name of ``struct rb_node`` member in entry type.
    :return: Iterator of ``type *`` objects.
    """
    type = root.prog_.type(type)
    for node in rbtree_preorder_for_each(root):
        yield container_of(node, type, member)


KeyType = TypeVar("KeyType")


def rb_find(
    type: Union[str, Type],
    root: Object,
    member: str,
    key: KeyType,
    cmp: Callable[[KeyType, Object], int],
) -> Object:
    """
    Find an entry in a red-black tree given a key and a comparator function.

    Note that this function does not have an analogue in the Linux kernel
    source code, as tree searches are all open-coded.

    :param type: Entry type.
    :param root: ``struct rb_root *``
    :param member: Name of ``struct rb_node`` member in entry type.
    :param key: Key to find.
    :param cmp: Callback taking key and entry that returns < 0 if the key is
        less than the entry, > 0 if the key is greater than the entry, and 0 if
        the key matches the entry.
    :return: ``type *`` found entry, or ``NULL`` if not found.
    """
    prog = root.prog_
    type = prog.type(type)
    node = root.rb_node.read_()
    while node:
        entry = container_of(node, type, member)
        ret = cmp(key, entry)
        if ret < 0:
            node = node.rb_left.read_()
        elif ret > 0:
            node = node.rb_right.read_()
        else:
            return entry
    return NULL(prog, prog.pointer_type(type))


def validate_rbtree(
    type: Union[str, Type],
    root: Object,
    member: str,
    cmp: Callable[[Object, Object], int],
    allow_equal: bool,
) -> None:
    """
    Validate a red-black tree.

    This checks that:

    1. The tree is a valid binary search tree ordered according to *cmp*.
    2. If *allow_equal* is ``False``, there are no nodes that compare equal
       according to *cmp*.
    3. The ``rb_parent`` pointers are consistent.
    4. The red-black tree requirements are satisfied: the root node is black,
       no red node has a red child, and every path from any node to any of its
       descendant leaf nodes goes through the same number of black nodes.

    :param type: Entry type.
    :param root: ``struct rb_root *``
    :param member: Name of ``struct rb_node`` member in entry type.
    :param cmp: Callback taking two ``type *`` entry objects that returns < 0
        if the first entry is less than the second entry, > 0 if the first
        entry is greater than the second entry, and 0 if they are equal.
    :param allow_equal: Whether the tree may contain entries that compare equal
        to each other.
    :raises ValidationError: if the tree is invalid
    """
    for _ in validate_rbtree_inorder_for_each_entry(
        type, root, member, cmp, allow_equal
    ):
        pass


def validate_rbtree_inorder_for_each_entry(
    type: Union[str, Type],
    root: Object,
    member: str,
    cmp: Callable[[Object, Object], int],
    allow_equal: bool,
) -> Iterator[Object]:
    """
    Like :func:`rbtree_inorder_for_each_entry()`, but validates the red-black
    tree like :func:`validate_rbtree()` while iterating.

    :param type: Entry type.
    :param root: ``struct rb_root *``
    :param member: Name of ``struct rb_node`` member in entry type.
    :param cmp: Callback taking two ``type *`` entry objects that returns < 0
        if the first entry is less than the second entry, > 0 if the first
        entry is greater than the second entry, and 0 if they are equal.
    :param allow_equal: Whether the tree may contain entries that compare equal
        to each other.
    :raises ValidationError: if the tree is invalid
    """

    prog = root.prog_
    type = prog.type(type)

    def visit(
        node: Object,
        parent_node: Object,
        parent_entry: Object,
        parent_is_red: bool,
        is_left: bool,
    ) -> Generator[Object, None, int]:
        if node:
            node_rb_parent, black = _rb_parent_color(node)
            if node_rb_parent != parent_node:
                raise ValidationError(
                    f"{parent_node.format_(dereference=False, symbolize=False)}"
                    f" rb_{'left' if is_left else 'right'}"
                    f" {node.format_(dereference=False, symbolize=False, type_name=False)}"
                    f" has rb_parent {node_rb_parent.format_(dereference=False, symbolize=False, type_name=False)}"
                )

            if parent_is_red and not black:
                raise ValidationError(
                    f"red node {parent_node.format_(dereference=False, symbolize=False)}"
                    f" has red child {node.format_(dereference=False, symbolize=False, type_name=False)}"
                )

            entry = container_of(node, type, member)
            r = cmp(entry, parent_entry)
            if r > 0:
                if is_left:
                    raise ValidationError(
                        f"{parent_entry.format_(dereference=False, symbolize=False)}"
                        f" left child {entry.format_(dereference=False, symbolize=False, type_name=False)}"
                        " compares greater than it"
                    )
            elif r < 0:
                if not is_left:
                    raise ValidationError(
                        f"{parent_entry.format_(dereference=False, symbolize=False)}"
                        f" right child {entry.format_(dereference=False, symbolize=False, type_name=False)}"
                        " compares less than it"
                    )
            elif not allow_equal:
                raise ValidationError(
                    f"{parent_entry.format_(dereference=False, symbolize=False)}"
                    f" {'left' if is_left else 'right'}"
                    f" child {entry.format_(dereference=False, symbolize=False, type_name=False)}"
                    " compares equal to it"
                )

            return (yield from descend(node, entry, black))
        else:
            return 0

    def descend(
        node: Object, entry: Object, black: bool
    ) -> Generator[Object, None, int]:
        left_black_height = yield from visit(
            node.rb_left.read_(), node, entry, parent_is_red=not black, is_left=True
        )
        yield entry
        right_black_height = yield from visit(
            node.rb_right.read_(), node, entry, parent_is_red=not black, is_left=False
        )
        if left_black_height != right_black_height:
            raise ValidationError(
                f"left and right subtrees of {node.format_(dereference=False, symbolize=False)}"
                f" have unequal black heights ({left_black_height} != {right_black_height})"
            )
        return left_black_height + black

    root_node = root.rb_node.read_()
    if root_node:
        parent, black = _rb_parent_color(root_node)
        if parent:
            raise ValidationError(
                f"root node {root_node.format_(dereference=False, symbolize=False)}"
                f" has parent {parent.format_(dereference=False, symbolize=False, type_name=False)}"
            )
        if not black:
            raise ValidationError(
                f"root node {root_node.format_(dereference=False, symbolize=False)} is red"
            )
        yield from descend(root_node, container_of(root_node, type, member), black)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash sym command."""

import argparse
import operator
import re
from typing import Any, List, Sequence

from drgn import Program, Symbol, SymbolKind
from drgn.commands import _repr_black, argument, drgn_argument, mutually_exclusive_group
from drgn.commands._crash.common import CrashDrgnCodeBuilder, crash_command

# Ignore symbols that the kernel excludes from kallsyms. See scripts/mksysmap
# in the Linux kernel source.
#
# scripts/mksysmap also ignores some symbols based on their nm(1) types, but we
# don't have a good way to do that.
_IGNORED_PREFIXES = (
    r"\$",
    r"\.L",
    r"__efistub_",
    r"__pi_\\$",
    r"__pi_\.L",
    r"__kvm_nvhe_\$",
    r"__kvm_nvhe_\.L",
    r"__\w*Thunk_",
    r"__kcfi_typeid_",
    r"__kvm_nvhe___kcfi_typeid_",
    r"__pi___kcfi_typeid_",
    r"__crc_",
    r"__kstrtab_",
    r"__kstrtabns_",
    r"__mod_device_table__",
)

_IGNORED_SUFFIXES = (
    r"_from_arm",
    r"_from_thumb",
    r"_veneer",
)

_IGNORED_EXACT = (
    r"L0",
    r"_SDA_BASE_",
    r"_SDA2_BASE_",
    r"__UNIQUE_ID_modinfo[0-9]*",
)

_IGNORED_CONTAINS = (
    r"\.long_branch\.",
    r"\.plt_branch\.",
)

_IGNORE_REGEX = re.compile(
    "|".join(
        [
            r"^(" + "|".join(_IGNORED_PREFIXES) + ")",
            r"(" + "|".join(_IGNORED_SUFFIXES) + ")$",
            r"^(" + "|".join(_IGNORED_EXACT) + ")$",
            r"(" + "|".join(_IGNORED_CONTAINS) + ")",
        ]
    )
)


def _include_symbol(sym: Symbol) -> bool:
    if sym.kind in {SymbolKind.SECTION, SymbolKind.FILE}:
        return False
    return not _IGNORE_REGEX.search(sym.name)


def _print_symbols(
    prog: Program, symbols: Sequence[Symbol], indent: str = "", verbose: bool = False
) -> None:
    for symbol in symbols:
        source_str = ""
        if verbose:
            try:
                source = prog.source_location(symbol.address)
            except LookupError:
                pass
            else:
                source_str = f" {source[0].filename}: {source[0].line}"
        print(f"{indent}{symbol.address:x} {symbol.name}{source_str}")


def _print_substring_matches(
    prog: Program, symbols: Sequence[Symbol], substring: str, indent: str = ""
) -> None:
    symbols = [symbol for symbol in symbols if substring in symbol.name]
    if symbols:
        _print_symbols(prog, symbols, indent)
    else:
        print(indent + "(none found)")


@crash_command(
    description="symbol table lookup",
    arguments=(
        mutually_exclusive_group(
            argument(
                "-l",
                dest="all",
                action="store_true",
                help="display all symbols and their values",
            ),
            argument(
                "-q",
                dest="string",
                action="append",
                help="""
                display all symbols whose name contains *STRING*. May be given
                multiple times
                """,
            ),
            # Crash allows -q to be combined with a name or address argument,
            # but that's confusing. Forbid it until someone complains.
            argument(
                "name_or_address",
                metavar="name|address",
                nargs="*",
                # Work around https://github.com/python/cpython/issues/72795
                # before Python 3.13.
                default=[],
                help="""
                symbol name or hexadecimal address to search for. May be given
                multiple times
                """,
            ),
            required=True,
        ),
        argument(
            "-p",
            dest="prev",
            action="store_true",
            help="display the target symbol and the previous symbol",
        ),
        argument(
            "-n",
            dest="next",
            action="store_true",
            help="display the target symbol and the next symbol",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_sym(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    if (args.next or args.prev) and not args.name_or_address:
        parser.error("-n/-p only make sense with name or address")

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)

        first = True
        if args.string or args.all or args.next or args.prev:
            code.add_import("operator")
            code.append(
                """\
all_symbols = prog.symbols()
all_symbols.sort(key=operator.attrgetter("address"))
"""
            )
            first = False

        if args.string is not None:
            code.append("\nfor sym in all_symbols:\n")
            if len(args.string) == 1:
                code.append(f"    if {_repr_black(args.string[0])} in sym.name:\n")
            else:
                code.append("    if any(substring in sym.name for substring in (")
                code.append(", ".join([_repr_black(string) for string in args.string]))
                code.append(")):\n")
            code.append("        ...\n")
            first = False

        for arg in args.name_or_address:
            if first:
                first = False
            else:
                code.append("\n")

            address = None
            try:
                prog.symbol(arg)
            except LookupError:
                try:
                    address = int(arg, 16)
                except ValueError:
                    pass

            if address is None:
                code.append(f"sym = prog.symbol({_repr_black(arg)})\n")
            else:
                code.append(f"sym = prog.symbol({hex(address)})\n")
            code.append(
                """\
try:
    source = prog.source_location(sym.address)
except LookupError:
    pass
"""
            )

            if args.prev or args.next:
                code.append("\ni = all_symbols.index(sym)\n")
            if args.prev:
                code.append(
                    """\
if i > 0:
    prev_sym = all_symbols[i - 1]
"""
                )
            if args.next:
                code.append(
                    """\
if i + 1 < len(all_symbols):
    next_sym = all_symbols[i + 1]
"""
                )

        if args.all:
            code.append(
                """\

for sym in all_symbols:
    ...
"""
            )

        return code.print()

    _all_symbols = None

    def all_symbols() -> List[Symbol]:
        nonlocal _all_symbols
        if _all_symbols is None:
            _all_symbols = [
                symbol for symbol in prog.symbols() if _include_symbol(symbol)
            ]
            _all_symbols.sort(key=operator.attrgetter("address"))
        return _all_symbols

    if args.string is not None:
        for string in args.string:
            _print_substring_matches(prog, all_symbols(), string)

    for arg in args.name_or_address:
        try:
            symbol = prog.symbol(arg)
        except LookupError:
            try:
                symbol = prog.symbol(int(arg, 16))
            except (ValueError, LookupError):
                print(f"symbol not found: {arg}\npossible alternatives:")
                _print_substring_matches(prog, all_symbols(), arg, "  ")
                continue

        symbols = [symbol]
        if args.prev or args.next:
            try:
                i = all_symbols().index(symbol)
            except ValueError:
                pass
            else:
                symbols = all_symbols()[
                    (i - 1 if args.prev and i > 0 else i) : (
                        i + 2 if args.next else i + 1
                    )
                ]

        _print_symbols(prog, symbols, verbose=True)

    if args.all:
        _print_symbols(prog, all_symbols())



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Programmable debugger

drgn is a programmable debugger. It is built on top of Python, so if you
don't know at least a little bit of Python, go learn it first.

drgn supports an interactive mode and a script mode. Both are simply a
Python interpreter initialized with a special drgn.Program object named
"prog" that represents the program which is being debugged.

In interactive mode, try

>>> help(prog)

or

>>> help(drgn.Program)

to learn more about how to use it.

Objects in the program (e.g., variables and values) are represented by
drgn.Object. Try

>>> help(drgn.Object)

Types are represented by drgn.Type objects. Try

>>> help(drgn.Type)

Various helpers are provided for particular types of programs. Try

>>> import drgn.helpers
>>> help(drgn.helpers)

The drgn.internal package contains the drgn internals. Everything in
that package should be considered implementation details and should not
be used.
"""

import io
import pkgutil
import sys
import types
from typing import Any, Dict, Optional, Tuple, Union, overload

from _drgn import (
    NULL,
    AbsenceReason,
    Architecture,
    DebugInfoOptions,
    ExtraModule,
    FaultError,
    FindObjectFlags,
    IntegerLike,
    KmodSearchMethod,
    Language,
    MainModule,
    MemorySearchIterator,
    MissingDebugInfoError,
    Module,
    ModuleFileStatus,
    NoDefaultProgramError,
    Object,
    ObjectAbsentError,
    ObjectNotFoundError,
    OutOfBoundsError,
    Path,
    Platform,
    PlatformFlags,
    PrimitiveType,
    Program,
    ProgramFlags,
    Qualifiers,
    Register,
    RelocatableModule,
    SharedLibraryModule,
    SourceLocation,
    SourceLocationList,
    StackFrame,
    StackTrace,
    SupplementaryFileKind,
    Symbol,
    SymbolBinding,
    SymbolIndex,
    SymbolKind,
    Thread,
    Type,
    TypeEnumerator,
    TypeKind,
    TypeKindSet,
    TypeMember,
    TypeParameter,
    TypeTemplateParameter,
    VdsoModule,
    WantedSupplementaryFile,
    alignof,
    cast,
    container_of,
    filename_matches,
    get_default_prog,
    host_platform,
    implicit_convert,
    offsetof,
    program_from_core_dump,
    program_from_kernel,
    program_from_pid,
    reinterpret,
    set_default_prog,
    sizeof,
)

# flake8 doesn't honor import X as X. See PyCQA/pyflakes#474.
# isort: split
from _drgn import (  # noqa: F401
    _elfutils_version as _elfutils_version,
    _enable_dlopen_debuginfod as _enable_dlopen_debuginfod,
    _have_debuginfod as _have_debuginfod,
    _with_json_c as _with_json_c,
    _with_libkdumpfile as _with_libkdumpfile,
    _with_lzma as _with_lzma,
    _with_pcre2 as _with_pcre2,
    _with_pcre2_utf as _with_pcre2_utf,
)
from drgn.internal.version import __version__ as __version__  # noqa: F401

__all__ = (
    "AbsenceReason",
    "Architecture",
    "DebugInfoOptions",
    "ExtraModule",
    "FaultError",
    "FindObjectFlags",
    "IntegerLike",
    "KmodSearchMethod",
    "Language",
    "MainModule",
    "MemorySearchIterator",
    "MissingDebugInfoError",
    "Module",
    "ModuleFileStatus",
    "NULL",
    "NoDefaultProgramError",
    "Object",
    "ObjectAbsentError",
    "ObjectNotFoundError",
    "OutOfBoundsError",
    "Path",
    "Platform",
    "PlatformFlags",
    "PrimitiveType",
    "Program",
    "ProgramFlags",
    "Qualifiers",
    "Register",
    "RelocatableModule",
    "SharedLibraryModule",
    "SourceLocation",
    "SourceLocationList",
    "StackFrame",
    "StackTrace",
    "SupplementaryFileKind",
    "Symbol",
    "SymbolBinding",
    "SymbolIndex",
    "SymbolKind",
    "Thread",
    "Type",
    "TypeEnumerator",
    "TypeKind",
    "TypeKindSet",
    "TypeMember",
    "TypeParameter",
    "TypeTemplateParameter",
    "VdsoModule",
    "WantedSupplementaryFile",
    "alignof",
    "cast",
    "container_of",
    "execscript",
    "filename_matches",
    "get_default_prog",
    "host_platform",
    "implicit_convert",
    "offsetof",
    "program_from_core_dump",
    "program_from_kernel",
    "program_from_pid",
    "reinterpret",
    "search_memory",
    "search_memory_regex",
    "search_memory_u16",
    "search_memory_u32",
    "search_memory_u64",
    "search_memory_word",
    "set_default_prog",
    "sizeof",
    "source_location",
    "stack_trace",
)


# From https://docs.python.org/3/reference/import.html#import-related-module-attributes.
_special_globals = frozenset(
    [
        "__name__",
        "__loader__",
        "__package__",
        "__spec__",
        "__path__",
        "__file__",
        "__cached__",
    ]
)


def execscript(path: str, *args: str, globals: Optional[Dict[str, Any]] = None) -> None:
    """
    Execute a script.

    The script is executed in the same context as the caller: currently defined
    globals are available to the script, and globals defined by the script are
    added back to the calling context.

    This is most useful for executing scripts from interactive mode. For
    example, you could have a script named ``exe.py``:

    .. code-block:: python3

        \"\"\"Get all tasks executing a given file.\"\"\"

        import sys

        from drgn.helpers.linux.fs import d_path
        from drgn.helpers.linux.pid import find_task

        def task_exe_path(task):
            if task.mm:
                return d_path(task.mm.exe_file.f_path).decode()
            else:
                return None

        tasks = [
            task for task in for_each_task()
            if task_exe_path(task) == sys.argv[1]
        ]

    Then, you could execute it and use the defined variables and functions:

    >>> execscript('exe.py', '/usr/bin/bash')
    >>> tasks[0].pid
    (pid_t)358442
    >>> task_exe_path(find_task(357954))
    '/usr/bin/vim'

    :param path: File path of the script.
    :param args: Zero or more additional arguments to pass to the script. This
        is a :ref:`variable argument list <python:tut-arbitraryargs>`.
    :param globals: If provided, globals to use instead of the caller's.
    """
    # This is based on runpy.run_path(), which we can't use because we want to
    # update globals even if the script throws an exception.
    saved_module = []
    try:
        saved_module.append(sys.modules["__main__"])
    except KeyError:
        pass
    saved_argv = sys.argv
    try:
        module = types.ModuleType("__main__")
        sys.modules["__main__"] = module
        sys.argv = [path]
        sys.argv.extend(args)

        with io.open_code(path) as f:
            code = pkgutil.read_code(f)
        if code is None:
            with io.open_code(path) as f:
                code = compile(f.read(), path, "exec")
        module.__spec__ = None
        module.__file__ = path
        module.__cached__ = None  # type: ignore[attr-defined]

        if globals is not None:
            caller_globals = globals
        else:
            caller_globals = sys._getframe(1).f_globals
        caller_special_globals = {
            name: caller_globals[name]
            for name in _special_globals
            if name in caller_globals
        }
        for name, value in caller_globals.items():
            if name not in _special_globals:
                setattr(module, name, value)

        try:
            exec(code, vars(module))
        finally:
            caller_globals.clear()
            caller_globals.update(caller_special_globals)
            for name, value in vars(module).items():
                if name not in _special_globals:
                    caller_globals[name] = value
    finally:
        sys.argv = saved_argv
        if saved_module:
            sys.modules["__main__"] = saved_module[0]
        else:
            del sys.modules["__main__"]


def stack_trace(thread: Union[Object, IntegerLike]) -> StackTrace:
    """
    Get the stack trace for the given thread using the :ref:`default program
    argument <default-program>`.

    See :meth:`Program.stack_trace()` for more details.

    :param thread: Thread ID, ``struct pt_regs`` object, or
        ``struct task_struct *`` object.
    """
    if isinstance(thread, Object):
        return thread.prog_.stack_trace(thread)
    else:
        return get_default_prog().stack_trace(thread)


def source_location(address: Union[IntegerLike, str], /) -> SourceLocationList:
    """
    Find the source code location containing a code address, similarly to
    :manpage:`addr2line(1)`, using the :ref:`default program argument
    <default-program>`.

    >>> source_location("__schedule")
    __schedule at kernel/sched/core.c:6646:1
    >>> source_location("__schedule+0x2b6")
    #0  context_switch at kernel/sched/core.c:5381:9
    #1  __schedule at kernel/sched/core.c:6765:8
    >>> source_location(0xffffffffb64d70a6)
    #0  context_switch at kernel/sched/core.c:5381:9
    #1  __schedule at kernel/sched/core.c:6765:8

    See :meth:`Program.source_location()` for more details.

    :param address: Code address as an integer, symbol name, or
        ``"symbol_name+offset"`` string.
    """
    if isinstance(address, Object):
        return address.prog_.source_location(address)
    else:
        return get_default_prog().source_location(address)


@overload
def search_memory(
    value: Union[bytes, str], *, alignment: int = 1
) -> MemorySearchIterator[int]:
    """
    Search for all non-overlapping occurrences of a byte string in the
    :ref:`default program's <default-program>` memory.

    .. code-block:: python3

        for address in search_memory(b"VMCOREINFO"):
            print(hex(address))

    :param value: Byte string to search for. If given as a :class:`str`, then
        this searches for its UTF-8 encoding.
    :param alignment: Only consider addresses aligned to this value (i.e.,
        ``address % alignment == 0``). Must be a power of two.
    :return: Iterator of addresses where the string is found.
    """
    ...


@overload
def search_memory(value: Union[IntegerLike, Object]) -> MemorySearchIterator[int]:
    """
    Search for all occurrences of a value in the :ref:`default program's
    <default-program>` memory.

    .. code-block:: python3

        ptr = stack_trace(pid)[2]["ptr"]
        for address in search_memory(ptr):
            print(hex(address))

    :param value: Value to search for. If given as an :class:`int`, then it is
        interpreted as a program word-sized, naturally aligned unsigned
        integer. If given as an :class:`Object`, then its size and alignment
        are determined from its type.
    :return: Iterator of addresses where the value is found.
    """
    ...


def search_memory(
    value: Union[bytes, str, IntegerLike, Object], **kwargs: Any
) -> MemorySearchIterator[int]:
    if isinstance(value, Object):
        return value.prog_.search_memory(value)
    else:
        return get_default_prog().search_memory(value)


def search_memory_u16(
    *values: Union[IntegerLike, Tuple[IntegerLike, IntegerLike]],
    ignore_mask: IntegerLike = 0,
) -> MemorySearchIterator[Tuple[int, int]]:
    """"""
    return get_default_prog().search_memory_u16(*values, ignore_mask=ignore_mask)


def search_memory_u32(
    *values: Union[IntegerLike, Tuple[IntegerLike, IntegerLike]],
    ignore_mask: IntegerLike = 0,
) -> MemorySearchIterator[Tuple[int, int]]:
    """"""
    return get_default_prog().search_memory_u32(*values, ignore_mask=ignore_mask)


def search_memory_u64(
    *values: Union[IntegerLike, Tuple[IntegerLike, IntegerLike]],
    ignore_mask: IntegerLike = 0,
) -> MemorySearchIterator[Tuple[int, int]]:
    """"""
    return get_default_prog().search_memory_u64(*values, ignore_mask=ignore_mask)


def search_memory_word(
    *values: Union[IntegerLike, Tuple[IntegerLike, IntegerLike]],
    ignore_mask: IntegerLike = 0,
) -> MemorySearchIterator[Tuple[int, int]]:
    """
    Search for all occurrences of one or more unsigned integers in the
    :ref:`default program's <default-program>` memory.

    :func:`search_memory_u16()`, :func:`search_memory_u32()`, and
    :func:`search_memory_u64()` search for 16-, 32- or 64-bit unsigned
    integers, respectively. :func:`search_memory_word()` searches for program
    word-sized unsigned integers. Natural alignment is assumed (i.e., 2-byte
    alignment for 16-bit integers, 4-byte alignment for 32-bit integers, 8-byte
    alignment for 64-bit integers).

    .. code-block:: python3

        for address, value in search_memory_word(
            0xdead000000000100, 0xdead000000000122
        ):
            print(hex(address), hex(value))

        for address, value in search_memory_word(
            0xdead000000000000, ignore_mask=0xffff
        ):
            print(hex(address), hex(value))

        obj = prog["obj"]
        for address, value in search_memory_word(
            (obj.address_, obj.address_ + sizeof(obj) - 1)
        ):
            print(hex(address), hex(value))

    :param values: Values to search for. Each value may be a single integer or
        a ``(min, max)`` range (where both ends are included).
    :param ignore_mask: Mask of bits to ignore when comparing to single
        integers given in *values*. This is not used when comparing to ranges.
    :return: Iterator of addresses where the value is found and the found
        value.
    """
    return get_default_prog().search_memory_word(*values, ignore_mask=ignore_mask)


@overload
def search_memory_regex(pattern: bytes) -> MemorySearchIterator[Tuple[int, bytes]]:
    """"""
    ...


@overload
def search_memory_regex(pattern: str) -> MemorySearchIterator[Tuple[int, str]]:
    r"""
    Search for all non-overlapping matches of a regular expression pattern in
    the program's memory.

    .. code-block:: python3

        # Search for anything that looks like root's password encrypted in
        # /etc/shadow.
        for address, match in search_memory_regex(rb"root:\$\w+\$[ -9;-~]+:"):
            print(hex(address), match)

    :param pattern: PCRE regular expression to search for. If given as
        :class:`bytes`, then search for 8-bit strings. If given as
        :class:`str`, then search for Unicode strings. Note that lookbehind
        assertions are not allowed.
    :return: Iterator of addresses where a match is found and the matching
        string.
    """
    ...


def search_memory_regex(
    pattern: Union[bytes, str]
) -> Union[
    MemorySearchIterator[Tuple[int, bytes]], MemorySearchIterator[Tuple[int, str]]
]:
    return get_default_prog().search_memory_regex(pattern)



#!/usr/bin/env drgn
# Copyright (c) 2025, Kylin Software, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import sys
from drgn import Object, FaultError, PlatformFlags, cast
from drgn.helpers.linux import find_task, follow_page, page_size
from drgn.helpers.linux.mm import (
    decode_page_flags, page_to_pfn, page_to_phys, page_to_virt, vma_find,
    PageSlab, PageCompound, PageHead, PageTail, compound_head, compound_order, compound_nr
)
from drgn.helpers.linux.cgroup import cgroup_name, cgroup_path

DESC = """
This is a drgn script to show the page state.
For more info on drgn, visit https://github.com/osandov/drgn.
"""

def format_page_data(page):
    """
    Format raw page data into a readable hex dump with "RAW:" prefix.

    :param page: drgn.Object instance representing the page.
    :return: Formatted string of memory contents.
    """
    try:
        address = page.value_()
        size = prog.type("struct page").size

        if prog.platform.flags & PlatformFlags.IS_64_BIT:
            word_size = 8
        else:
            word_size = 4
        num_words = size // word_size

        values = []
        for i in range(num_words):
            word_address = address + i * word_size
            word = prog.read_word(word_address)
            values.append(f"{word:0{word_size * 2}x}")

        lines = [f"RAW: {' '.join(values[i:i + 4])}" for i in range(0, len(values), 4)]

        return "\n".join(lines)

    except FaultError as e:
        return f"Error reading memory: {e}"
    except Exception as e:
        return f"Unexpected error: {e}"

def get_memcg_info(page):
    """Retrieve memory cgroup information for a page."""
    try:
        MEMCG_DATA_OBJEXTS = prog.constant("MEMCG_DATA_OBJEXTS").value_()
        MEMCG_DATA_KMEM = prog.constant("MEMCG_DATA_KMEM").value_()
        mask = prog.constant('__NR_MEMCG_DATA_FLAGS').value_() - 1
        memcg_data = page.memcg_data.read_()
        if memcg_data & MEMCG_DATA_OBJEXTS:
            slabobj_ext = cast("struct slabobj_ext *", memcg_data & ~mask)
            memcg = slabobj_ext.objcg.memcg.value_()
        elif memcg_data & MEMCG_DATA_KMEM:
            objcg = cast("struct obj_cgroup *", memcg_data & ~mask)
            memcg = objcg.memcg.value_()
        else:
            memcg = cast("struct mem_cgroup *", memcg_data & ~mask)

        if memcg.value_() == 0:
            return "none", "/sys/fs/cgroup/memory/"
        cgrp = memcg.css.cgroup
        return cgroup_name(cgrp).decode(), f"/sys/fs/cgroup/memory{cgroup_path(cgrp).decode()}"
    except FaultError as e:
        return "unknown", f"Error retrieving memcg info: {e}"
    except Exception as e:
        return "unknown", f"Unexpected error: {e}"

def show_page_state(page, addr, mm, pid, task):
    """Display detailed information about a page."""
    try:
        print(f'PID: {pid} Comm: {task.comm.string_().decode()} mm: {hex(mm)}')
        try:
            print(format_page_data(page))
        except FaultError as e:
            print(f"Error reading page data: {e}")
        fields = {
            "Page Address": hex(page.value_()),
            "Page Flags": decode_page_flags(page),
            "Page Size": prog["PAGE_SIZE"].value_(),
            "Page PFN": hex(page_to_pfn(page).value_()),
            "Page Physical": hex(page_to_phys(page).value_()),
            "Page Virtual": hex(page_to_virt(page).value_()),
            "Page Refcount": page._refcount.counter.value_(),
            "Page Mapcount": page._mapcount.counter.value_(),
            "Page Index": hex(page.__folio_index.value_()),
            "Page Memcg Data": hex(page.memcg_data.value_()),
        }

        memcg_name, memcg_path = get_memcg_info(page)
        fields["Memcg Name"] = memcg_name
        fields["Memcg Path"] = memcg_path
        fields["Page Mapping"] = hex(page.mapping.value_())
        fields["Page Anon/File"] = "Anon" if page.mapping.value_() & 0x1 else "File"

        try:
            vma = vma_find(mm, addr)
            fields["Page VMA"] = hex(vma.value_())
            fields["VMA Start"] = hex(vma.vm_start.value_())
            fields["VMA End"] = hex(vma.vm_end.value_())
        except FaultError as e:
            fields["Page VMA"] = "Unavailable"
            fields["VMA Start"] = "Unavailable"
            fields["VMA End"] = "Unavailable"
            print(f"Error retrieving VMA information: {e}")

        # Calculate the maximum field name length for alignment
        max_field_len = max(len(field) for field in fields)

        # Print aligned fields
        for field, value in fields.items():
            print(f"{field}:".ljust(max_field_len + 2) + f"{value}")

        # Additional information about the page
        if PageSlab(page):
            print("This page belongs to the slab allocator.")

        if PageCompound(page):
            print("This page is part of a compound page.")
            if PageHead(page):
                print("This page is the head page of a compound page.")
            if PageTail(page):
                print("This page is the tail page of a compound page.")
            print(f"{'Head Page:'.ljust(max_field_len + 2)}{hex(compound_head(page).value_())}")
            print(f"{'Compound Order:'.ljust(max_field_len + 2)}{compound_order(page).value_()}")
            print(f"{'Number of Pages:'.ljust(max_field_len + 2)}{compound_nr(page).value_()}")
        else:
            print("This page is not part of a compound page.")
    except FaultError as e:
        print(f"Error accessing page state: {e}")
    except Exception as e:
        print(f"Unexpected error: {e}")

def main():
    """Main function to parse arguments and display page state."""
    parser = argparse.ArgumentParser(description=DESC, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('pid', metavar='PID', type=int, help='Target process ID (PID)')
    parser.add_argument('vaddr', metavar='VADDR', type=str, help='Target virtual address in hexadecimal format (e.g., 0x7fff1234abcd)')
    args = parser.parse_args()

    try:
        vaddr = int(args.vaddr, 16)
    except ValueError:
        sys.exit(f"Error: Invalid virtual address format: {args.vaddr}")

    try:
        task = find_task(args.pid)
        mm = task.mm
        page = follow_page(mm, vaddr)

        if page:
            show_page_state(page, vaddr, mm, args.pid, task)
        else:
            sys.exit(f"Address {hex(vaddr)} is not mapped.")
    except FaultError as e:
        sys.exit(f"Error accessing task or memory: {e}")
    except Exception as e:
        sys.exit(f"Unexpected error: {e}")

if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

import argparse
import re
from typing import Any, List, Literal, Tuple, Union

from drgn import FaultError, MemorySearchIterator, Program
from drgn.commands import (
    _repr_black,
    _repr_raw_bytes,
    argument,
    argument_group,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash._rd import _print_memory
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _addr_or_sym,
    _resolve_addr_or_sym,
    crash_command,
    print_task_header,
)
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.pid import for_each_task
from drgn.helpers.linux.sched import cpu_curr

# Number of bytes to print for a string search (including the matching string
# itself). This is copied from crash (although crash seems to truncate the
# searched string to this size, too).
_STRING_CONTEXT = 56


def _search_drgn_option(prog: Program, args: argparse.Namespace) -> None:
    code = CrashDrgnCodeBuilder(prog)

    for_targets = (
        "address" if args.strings and len(args.values) == 1 else "address, value"
    )

    need_it = (
        args.task_stacks
        or args.active_task_stacks
        or args.start is not None
        or args.end is not None
        or args.length is not None
        or args.physical
    )

    if need_it:
        code.append("it = ")
    else:
        code.append(f"for {for_targets} in ")

    if args.strings:
        if len(args.values) == 1:
            code.add_from_import("drgn", "search_memory")
            code.append("search_memory(")
            code.append(_repr_black(args.values[0]))
            code.append(")")
        else:
            pattern = bytearray()
            for value in args.values:
                if pattern:
                    pattern.extend(b"|")
                pattern.extend(re.escape(value.encode()))
            code.add_from_import("drgn", "search_memory_regex")
            code.append("search_memory_regex(")
            code.append(_repr_raw_bytes(pattern))
            code.append(")")
    else:
        if args.u32:
            iterator_func = "search_memory_u32"
        elif args.u16:
            iterator_func = "search_memory_u16"
        else:
            iterator_func = "search_memory_word"
        code.add_from_import("drgn", iterator_func)
        code.append(iterator_func)
        code.append("(")

        first = True
        for value in args.values:
            if first:
                first = False
            else:
                code.append(", ")
            code.append_addr_or_sym(_addr_or_sym(value))

        if args.ignore_mask is not None:
            code.append(f", ignore_mask={args.ignore_mask:#x}")
        code.append(")")

    if need_it:
        code.append("\n")
    else:
        code.begin_block(":\n")

    if args.task_stacks or args.active_task_stacks:
        code.append('\nstack_size = prog["THREAD_SIZE"].value_()\n')
        if args.task_stacks:
            code.add_from_import("drgn.helpers.linux.pid", "for_each_task")
            code.begin_block("for task in for_each_task(idle=True):\n")
        else:
            code.add_from_import("drgn.helpers.linux.cpumask", "for_each_online_cpu")
            code.add_from_import("drgn.helpers.linux.sched", "cpu_curr")
            code.begin_block("for cpu in for_each_online_cpu():\n")
            code.append("task = cpu_curr(cpu)\n")
        code.append(
            """\
stack = task.stack.value_()
if not stack:
    continue
it.set_address_range(stack, stack + stack_size - 1)
"""
        )

    if (
        args.start is not None
        or args.end is not None
        or args.length is not None
        or args.physical
    ):
        first = True
        if args.start is not None and args.length is not None:
            code.append("min_address = ")
            code.append_addr_or_sym(args.start)
            code.append("\n")
            code.append(
                f"it.set_address_range(min_address=min_address, max_address=min_address + {args.length} - 1"
            )
            first = False
        else:
            code.append("it.set_address_range(")
            first = True
            if args.start is not None:
                code.append("min_address=")
                code.append_addr_or_sym(args.start)
                first = False
            if args.end is not None:
                if first:
                    first = False
                else:
                    code.append(", ")
                code.append("max_address=")
                code.append_addr_or_sym(args.end)
                code.append(" - 1")
            elif args.length is not None:
                assert first
                code.append(", ")
                first = False
                code.append(f"max_address = {args.length} - 1")

        if args.physical:
            if first:
                first = False
            else:
                code.append(", ")
            code.append("physical=True")
        code.append(")\n")

    if need_it:
        code.begin_block(f"for {for_targets} in it:\n")
    code.append("...\n")

    code.end_block()  # For the it loop.
    if args.task_stacks or args.active_task_stacks:
        code.end_block()

    code.print()


@crash_command(
    description="search memory",
    long_description="""
    Display every memory address where one or more values are found.
    """,
    arguments=(
        mutually_exclusive_group(
            argument(
                "-p",
                dest="physical",
                action="store_true",
                help="search physical addresses",
            ),
            argument(
                "-t",
                dest="task_stacks",
                action="store_true",
                help="search only kernel stacks of all tasks",
            ),
            argument(
                "-T",
                dest="active_task_stacks",
                action="store_true",
                help="search only kernel stacks of active tasks",
            ),
        ),
        argument(
            "-s",
            dest="start",
            type="addr_or_sym",
            help="start searching at this hexadecimal address or symbol name",
        ),
        mutually_exclusive_group(
            argument(
                "-e",
                dest="end",
                type="addr_or_sym",
                help="stop searching at this hexadecimal address or symbol name",
            ),
            argument(
                "-l",
                dest="length",
                type="decimal_or_hexadecimal",
                help="search this many bytes",
            ),
        ),
        argument_group(
            mutually_exclusive_group(
                argument(
                    "-c",
                    dest="strings",
                    action="store_true",
                    help="search for strings",
                ),
                argument(
                    "-w",
                    dest="u32",
                    action="store_true",
                    help="search for unsigned ints",
                ),
                argument(
                    "-h",
                    dest="u16",
                    action="store_true",
                    help="search for unsigned shorts",
                ),
            ),
            title="units",
            description="""
            What unit the values to search for are in. The default is unsigned
            long.
            """,
        ),
        argument(
            "-x",
            dest="context",
            type=int,
            help="""
            display this many extra units of memory before and after every
            found value. Cannot be used with **-c**
            """,
        ),
        argument(
            "-m",
            dest="ignore_mask",
            type="hexadecimal",
            help="""
            ignore bits set in this hexadecimal mask. Cannot be used with
            **-c**
            """,
        ),
        argument(
            "values",
            metavar="value",
            nargs="+",
            help="""
            value to search for. Unless **-c** is given, then this is a
            hexadecimal integer, or a symbol name to use its address. If **-c**
            is given, then this is an arbitrary string
            """,
        ),
        drgn_argument,
    ),
)
def _crash_cmd_search(
    prog: Program,
    name: str,
    args: argparse.Namespace,
    *,
    parser: argparse.ArgumentParser,
    **kwargs: Any,
) -> None:
    if args.task_stacks or args.active_task_stacks:
        if args.start is not None:
            parser.error("-s cannot be used with -t or -T")
        if args.end is not None:
            parser.error("-e cannot be used with -t or -T")
        if args.length is not None:
            parser.error("-l cannot be used with -t or -T")

    if args.strings:
        if args.ignore_mask is not None:
            parser.error("-m cannot be used with -c")
        if args.context is not None:
            parser.error("-x cannot be used with -c")

    if args.drgn:
        return _search_drgn_option(prog, args)

    address_size = prog.address_size()
    symbols = {}
    it: Union[
        MemorySearchIterator[Tuple[int, int]], MemorySearchIterator[Tuple[int, bytes]]
    ]
    if args.strings:
        pattern_parts: List[bytes] = []
        for value in args.values:
            value_bytes = value.encode()
            if pattern_parts:
                pattern_parts.append(b"|")
            pattern_parts.append(re.escape(value_bytes))
            if len(value_bytes) < _STRING_CONTEXT:
                pattern_parts.append(b"(?s:.){0,")
                pattern_parts.append(str(_STRING_CONTEXT - len(value_bytes)).encode())
                pattern_parts.append(b"}")
        it = prog.search_memory_regex(b"".join(pattern_parts))
    else:
        values = []
        for value in args.values:
            try:
                values.append(int(value, 16))
            except ValueError:
                symbol_name = value
                value = prog.symbol(symbol_name).address
                values.append(value)
                symbols[value] = symbol_name
        if args.ignore_mask is None:
            args.ignore_mask = 0
        if args.u32:
            unit: Literal[1, 2, 4, 8] = 4
            it = prog.search_memory_u32(*values, ignore_mask=args.ignore_mask)
        elif args.u16:
            unit = 2
            it = prog.search_memory_u16(*values, ignore_mask=args.ignore_mask)
        else:
            unit = address_size  # type: ignore[assignment]
            it = prog.search_memory_word(*values, ignore_mask=args.ignore_mask)

    first_match = True

    def print_match(address: int, value: Union[int, bytes]) -> None:
        nonlocal first_match

        if isinstance(value, int):
            value_str = f"{value:x}"
            try:
                symbol_name = symbols[value]
            except KeyError:
                pass
            else:
                value_str = f"{value_str} ({symbol_name})"
        else:
            # Replace non-printable characters with ".".
            value_str = re.sub(rb"[^ -~]", b".", value).decode("ascii")

        if args.context:
            if first_match:
                first_match = False
            else:
                print()

            for size in range(args.context * unit, 0, -unit):
                try:
                    mem = prog.read(address - size, size)
                except FaultError:
                    pass
                else:
                    _print_memory(prog, address - size, mem, unit)
                    break

            print(f"{address:{address_size * 2}x}:  {value_str}")

            for size in range(args.context * unit, 0, -unit):
                try:
                    mem = prog.read(address + unit, size)
                except FaultError:
                    pass
                else:
                    _print_memory(prog, address + unit, mem, unit)
                    break
        else:
            print(f"{address:x}: {value_str}")

    if args.task_stacks or args.active_task_stacks:
        if args.task_stacks:
            tasks = for_each_task(prog, idle=True)
        else:
            tasks = (cpu_curr(prog, cpu) for cpu in for_each_online_cpu(prog))
        first = True
        thread_size = prog["THREAD_SIZE"].value_()
        for task in tasks:
            stack = task.stack.value_()
            if not stack:
                continue
            it.set_address_range(stack, stack + thread_size - 1)
            found_match = False
            for address, value in it:
                if not found_match:
                    if first:
                        first = False
                    else:
                        print()
                    print_task_header(task)
                    found_match = True
                print_match(address, value)
    else:
        min_address = 0
        max_address = None
        if args.start is not None:
            min_address = _resolve_addr_or_sym(prog, args.start)
        if args.end is not None:
            max_address = _resolve_addr_or_sym(prog, args.end) - 1
        elif args.length is not None:
            max_address = min_address + args.length - 1
        it.set_address_range(
            min_address=min_address, max_address=max_address, physical=args.physical
        )
        for address, value in it:
            print_match(address, value)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Devices
-------

The ``drgn.helpers.linux.device`` module provides helpers for working with
Linux devices, including the kernel encoding of ``dev_t``.
"""

import operator
from typing import Iterable, Iterator, Tuple

from drgn import NULL, IntegerLike, Object, Program, cast
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import list_for_each_entry

__all__ = (
    "MAJOR",
    "MINOR",
    "MKDEV",
    "bus_for_each_dev",
    "bus_to_subsys",
    "class_for_each_device",
    "class_to_subsys",
    "dev_name",
    "for_each_registered_blkdev",
    "for_each_registered_chrdev",
)


# This hasn't changed since at least v2.6.
_MINORBITS = 20
_MINORMASK = (1 << _MINORBITS) - 1


def MAJOR(dev: IntegerLike) -> int:
    """
    Return the major ID of a kernel ``dev_t``.

    :param dev: ``dev_t`` object or :class:`int`.
    """
    return operator.index(dev) >> _MINORBITS


def MINOR(dev: IntegerLike) -> int:
    """
    Return the minor ID of a kernel ``dev_t``.

    :param dev: ``dev_t`` object or :class:`int`.
    """
    return operator.index(dev) & _MINORMASK


def MKDEV(major: IntegerLike, minor: IntegerLike) -> int:
    """
    Return a kernel ``dev_t`` from the major and minor IDs.

    :param major: Device major ID.
    :param minor: Device minor ID.
    """
    return (operator.index(major) << _MINORBITS) | operator.index(minor)


def dev_name(dev: Object) -> bytes:
    """
    Get the name of a device.

    :param dev: ``struct device *``
    """
    init_name = dev.init_name.read_()
    if init_name:
        return init_name.string_()
    return dev.kobj.name.string_()


def bus_to_subsys(bus: Object) -> Object:
    """
    Get the private data for a device bus.

    :param bus: ``struct bus_type *``
    :return: ``struct subsys_private *``
    """
    prog = bus.prog_
    # Walk the list of registered busses to find the struct subsys_private
    # matching the given bus. Note that before Linux kernel commit d2bf38c088e0
    # ("driver core: remove private pointer from struct bus_type") (in v6.3),
    # struct subsys_private could also be found in struct bus::p, but it's
    # easier to only maintain the newer code path.
    for sp in list_for_each_entry(
        "struct subsys_private",
        prog["bus_kset"].list.address_of_(),
        "subsys.kobj.entry",
    ):
        if sp.bus == bus:
            return sp
    return NULL(prog, "struct subsys_private *")


def bus_for_each_dev(bus: Object) -> Iterable[Object]:
    """
    Iterate over all devices on a bus.

    :param bus: ``struct bus_type *``
    :return: Iterator of ``struct device *`` objects.
    """
    sp = bus_to_subsys(bus)
    for dev_prv in list_for_each_entry(
        "struct device_private",
        sp.klist_devices.k_list.address_of_(),
        "knode_bus.n_node",
    ):
        yield dev_prv.device.read_()


def class_to_subsys(class_: Object) -> Object:
    """
    Get the private data for a device class.

    :param bus: ``struct class *``
    :return: ``struct subsys_private *``
    """
    prog = class_.prog_
    # Walk the list of registered classes to find the struct subsys_private
    # matching the given class. Note that before Linux kernel commit
    # 2df418cf4b72 ("driver core: class: remove subsystem private pointer from
    # struct class") (in v6.4), struct subsys_private could also be found in
    # struct class::p, but it's easier to only maintain the newer code path.
    for sp in list_for_each_entry(
        "struct subsys_private",
        prog["class_kset"].list.address_of_(),
        "subsys.kobj.entry",
    ):
        if sp.member_("class") == class_:
            return sp
    return NULL(prog, "struct subsys_private *")


# The naming inconsistency between this and bus_for_each_dev() is inherited
# from the kernel source code :(
def class_for_each_device(class_: Object) -> Iterable[Object]:
    """
    Iterate over all devices of a class.

    :param bus: ``struct class *``
    :return: Iterator of ``struct device *`` objects.
    """
    prog = class_.prog_
    try:
        class_in_device_private = prog.cache["class_in_device_private"]
    except KeyError:
        # Linux kernel commit 570d0200123f ("driver core: move
        # device->knode_class to device_private") (in v5.1) moved the list
        # node.
        class_in_device_private = prog.type("struct device_private").has_member(
            "knode_class"
        )
        prog.cache["class_in_device_private"] = class_in_device_private

    devices = class_to_subsys(class_).klist_devices.k_list.address_of_()
    if class_in_device_private:
        for dev_prv in list_for_each_entry(
            "struct device_private", devices, "knode_class.n_node"
        ):
            yield dev_prv.device.read_()
    else:
        yield from list_for_each_entry("struct device", devices, "knode_class.n_node")


@takes_program_or_default
def for_each_registered_chrdev(
    prog: Program,
) -> Iterator[Tuple[int, int, bytes, Object]]:
    """
    Iterate over all registered character device number ranges.

    Character device numbers are reserved in ranges, so this returns the first
    ``dev_t`` value (which includes a major and minor ID) and the size of the
    range (i.e., the number of consecutive minor IDs starting from the first
    ``dev_t``).

    :return: Iterator of (first device number, number of consecutive minor IDs,
        name, ``struct cdev *`` object) tuples. The ``struct cdev *`` may be
        ``NULL``.
    """
    cdev_map_probes = prog["cdev_map"].probes
    for cd in prog["chrdevs"]:
        while cd := cd.read_():
            major = cd.major.value_()
            dev = MKDEV(major, cd.baseminor)
            cdev = cd.cdev.read_()
            if not cdev:
                probe = cdev_map_probes[major].read_()
                while next := probe.next.read_():
                    if probe.dev.value_() == dev:
                        cdev = cast(cdev.type_, probe.data)
                        break
                    probe = next

            yield dev, cd.minorct.value_(), cd.name.string_(), cdev
            cd = cd.next


@takes_program_or_default
def for_each_registered_blkdev(prog: Program) -> Iterator[Tuple[int, bytes, Object]]:
    """
    Iterate over all registered block device numbers.

    Block device numbers are reserved by major ID.

    :return: Iterator of (major ID, name, ``struct blk_major_name *`` object)
        tuples.
    """
    for name in prog["major_names"]:
        while name := name.read_():
            yield name.major.value_(), name.name.string_(), name
            name = name.next



# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later
"""
qemu_mem.py - speed up memory access for QEMU programs via /proc/pid/mem

Parsing the QEMU HMP "xp" command is much slower than native reads for multiple
reasons. While that's great to boostrap a debugging session, it may be possible
to speed up the process. If QEMU is creating a single, anonymous memory region
for guest memory, we may be able to create a faster reader via /proc/pid/mem.
While this is not a common configuration for real-world virtualization, it's a
good starting point and could probably be extended to more practical use cases.

Use this in one of two ways:

1. Once drgn has started: execscript("path/to/qemu_mem.py")
2. As a drgn plugin: DRGN_PLUGINS=qemu:path/to/qemu_mem.py drgn ...
"""
import logging
import os
import socket
import struct
from pathlib import Path

from _drgn import _linux_helper_follow_phys
from drgn import Program, ProgramFlags


log = logging.getLogger("drgn.qemu_mem")


def find_qemu_pid_direct():
    """
    Assuming that we're connected via a Unix socket, find the first socket FD we
    have open, and then get the PID associated with it. This is very likely to
    be the QEMU PID we're talking to, but like everything else in this file,
    it's a heuristic.
    """
    for p in Path("/proc/self/fd").iterdir():
        if p.readlink().name.startswith("socket:"):
            fd = int(p.name)
            break
    else:
        return None

    try:
        s = socket.socket(fileno=fd)
        # Ironically, we could omit the buffer size, and Python would treat this
        # as an integer socket option. Since what we care about is the first
        # integer of the structure, the return value would actually just be the
        # PID we care about! But that's super hacky, plus a PID should be
        # *unsigned*, so let's do it the correct way.
        fmt = "=III"
        buf = s.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, struct.calcsize(fmt))
        result = struct.unpack(fmt, buf)
        return result[0]
    except OSError:
        return None
    finally:
        s.detach()


def find_qemu_pids():
    """
    Using the naive assumption that qemu processes (and only qemu processes)
    have commands whose names start with "qemu-", find all candidate QEMU pids.
    NB: drgn might be able to get a specific PID from the Unix socket with
    getsockopt(SO_PEERCRED), which would be vastly preferable to this approach.
    """
    pid = find_qemu_pid_direct()
    if pid is not None:
        return [pid], "sockopt"
    pids = []
    for d in Path("/proc").iterdir():
        try:
            if (d / "comm").open("rb").read().startswith(b"qemu"):
                pids.append(int(d.name))
        except OSError:
            pass
    return pids, "procfs"


def find_matching_phys_map(pid, fd, offset, data):
    """
    Using the naive assumption that qemu processes map guest physical memory in
    one contiguous anonymous VMA, find the base address of this mapping. The
    caller provides some data at a known offset, which we use to help identify
    which anonymous mapping is correct.
    """
    maps = Path(f"/proc/{pid}/maps").open("rt").read()

    min_len = offset + len(data)
    for line in maps.splitlines():
        line = line.strip()
        if line:
            fields = line.split()
            if len(fields) > 5:
                continue
            start, end = fields[0].split("-")
            start, end = int(start, 16), int(end, 16)
            vma_size = end - start
            if vma_size < min_len:
                continue
            try:
                if os.pread(fd, len(data), start + offset) == data:
                    return start, end
            except OSError:
                continue
    log.debug("error: could not find QEMU physical VMA")
    return None


def main(prog: Program):
    # Find the QEMU pid we're targeting, hopefully
    pids, meth = find_qemu_pids()
    if len(pids) != 1:
        log.debug(f"error: looking for one qemu pid via {meth}, found {len(pids)}")
        return
    pid = pids[0]

    # Open its memory if possible
    try:
        fd = os.open(f"/proc/{pid}/mem", os.O_RDONLY)
    except PermissionError:
        log.debug("error: No permission to open QEMU /proc/pid/mem")
        return

    # We need a valid physical address with real data to ensure we pick the
    # right VMA. Drgn has workarounds for both x86_64 and aarch64 that allow us
    # to map virtual to physical addresses without debuginfo or virtual
    # mappings. Other architectures may have those in the future.
    pfx = "SYMBOL(swapper_pg_dir)="
    for line in prog["VMCOREINFO"].string_().decode().splitlines():
        if line.startswith(pfx):
            test_vaddr = int(line[len(pfx):], 16)
            break
    else:
        log.debug("error: could not find init_uts_ns symbol in VMCOREINFO")
        return
    test_paddr = _linux_helper_follow_phys(prog, test_vaddr, test_vaddr)
    expected = prog.read(test_paddr, 2 * prog["PAGE_SIZE"].value_(), True)

    vma = find_matching_phys_map(pid, fd, test_paddr, expected)
    if vma is None:
        return

    log.info(f"found matching VMA @ {vma[0]:x} for QEMU pid {pid} (via {meth}), mapping...")

    def reader(a, c, o, p):
        return os.pread(fd, c, o+vma[0])

    prog.add_memory_segment(0, 0xFFFFFFFFFFFFFFFF, reader, True)


def drgn_prog_set(prog: Program):
    if prog.flags & (
        ProgramFlags.IS_LINUX_KERNEL | ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL
    ) == (ProgramFlags.IS_LINUX_KERNEL | ProgramFlags.IS_LIVE):
        main(prog)


if __name__ == "__main__":
    main(prog)  # noqa



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash commands related to the current context."""

import argparse
from typing import Any, Callable, NamedTuple, Optional

from drgn import Object, Program
from drgn.commands import (
    CommandArgumentError,
    CommandError,
    argument,
    drgn_argument,
    mutually_exclusive_group,
)
from drgn.commands._crash._sys import _append_sys_drgn_task, _print_sys
from drgn.commands._crash.common import (
    CrashDrgnCodeBuilder,
    _crash_foreach_subcommand,
    _crash_get_panic_context,
    _find_pager,
    _get_pager,
    _pid_or_task,
    _TaskSelector,
    crash_command,
    crash_get_context,
    print_task_header,
)
from drgn.helpers.linux.sched import cpu_curr


def _show_scroll_option(prog: Program) -> None:
    on = "on" if prog.config.get("crash_scroll", True) else "off"
    pager = _get_pager(prog)
    if pager:
        print(f"scroll: {on} ({pager})")
    else:
        print("scroll: off (pager not found)")


def _validate_scroll_option(value: str) -> None:
    if value not in {"on", "off", "less", "more"}:
        raise CommandArgumentError(
            f"set: error: invalid value for scroll: {value!r} "
            "(must be on, off, less, or more)"
        )


def _set_scroll_option(prog: Program, value: str) -> None:
    if value == "on":
        prog.config["crash_scroll"] = True
    elif value == "off":
        prog.config["crash_scroll"] = False
    else:
        pager = _find_pager(value)
        if not pager:
            raise CommandError("pager not found")
        prog.config["crash_pager"] = pager


def _show_radix_option(prog: Program) -> None:
    radix = prog.config.get("crash_radix", 10)
    print(f"output radix: {radix} ({'hex' if radix == 16 else 'decimal'})")


_RADICES = {
    # Crash allows anything starting with "dec", "ten", "hex", or "six". We are
    # stricter.
    "10": 10,
    "dec": 10,
    "decimal": 10,
    "ten": 10,
    "16": 16,
    "hex": 16,
    "hexadecimal": 16,
    "sixteen": 16,
}


def _validate_radix_option(value: str) -> None:
    if value not in _RADICES:
        raise CommandArgumentError(
            f"set: error: invalid value for radix: {value!r} (must be 10 or 16)"
        )


def _set_radix_option(prog: Program, value: str) -> None:
    prog.config["crash_radix"] = _RADICES[value]


class _CrashOption(NamedTuple):
    show: Callable[[Program], None]
    validate: Callable[[str], None]
    set: Callable[[Program, str], None]


_OPTIONS = {
    "scroll": _CrashOption(
        _show_scroll_option, _validate_scroll_option, _set_scroll_option
    ),
    "radix": _CrashOption(
        _show_radix_option, _validate_radix_option, _set_radix_option
    ),
}


@crash_command(
    description="set current context or configuration",
    long_description="""
    Set/show the default context used by commands that target a task, or
    set/show configuration.

    If no options are given, then the current context is displayed.

    If given a configuration option name and a value, then the option is set to
    that value. If given a configuration option name only, then the current
    value is shown. The following options and values are supported:

    * ``scroll on | off``: enable (the default) or disable scrolling of long output.

    * ``scroll less | more``: set the pager program. The default is ``less``.

    * ``radix 10 | 16``: set the default integer output base. The default is 10.
    """,
    usage=r"**set** [*pid* | *task* | **-p** | **-c** *CPU* | *option* [*value*]] [**\-\-drgn**]",
    arguments=(
        mutually_exclusive_group(
            # argparse can't express pid | task | option [value], so we use
            # this hack to consume all of the positional arguments (which we
            # parse manually) and define dummy arguments that are only used for
            # the help string.
            argument(
                "posargs",
                metavar="pid_or_task_or_option",
                nargs="*",
                # Work around https://github.com/python/cpython/issues/72795
                # before Python 3.13.
                default=[],
                help=argparse.SUPPRESS,
            ),
            argument(
                "task",
                metavar="pid",
                nargs="?",
                help="set the current context to the task "
                "with the given decimal process ID",
            ),
            argument(
                "task",
                nargs="?",
                help="set the current context to the task "
                "with the hexadecimal ``task_struct`` address",
            ),
            argument(
                "-c",
                dest="cpu",
                type=int,
                help="set the current context to the active task on the given CPU",
            ),
            argument(
                "-p",
                dest="panic",
                action="store_true",
                help="set the current context to the task that crashed (for core dumps) "
                "or the drgn process itself (for live kernels)",
            ),
            argument("option", nargs="?", help="option to get or set"),
            argument("value", nargs="?", help="value to set option to"),
        ),
        drgn_argument,
    ),
)
def _crash_cmd_set(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> Optional[Object]:
    if args.posargs:
        option = _OPTIONS.get(args.posargs[0])
        if option is not None:
            if args.drgn:
                raise CommandError("no drgn equivalent")

            if len(args.posargs) > 1:
                option.validate(args.posargs[1])

                if len(args.posargs) > 2:
                    raise CommandArgumentError(
                        "set: error: unrecognized arguments: "
                        + " ".join(args.posargs[2:])
                    )

                option.set(prog, args.posargs[1])

            option.show(prog)
            return None

        try:
            args.task = _pid_or_task(args.posargs[0])
        except ValueError:
            raise CommandArgumentError(
                f"set: error: not a pid, task, or option: {args.posargs[0]!r}"
            )

        if len(args.posargs) > 1:
            raise CommandArgumentError(
                "set: error: unrecognized arguments: " + " ".join(args.posargs[1:])
            )

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        if args.panic:
            code._append_crash_panic_context()
        elif args.cpu is not None:
            code._append_crash_cpu_context(args.cpu)
        elif args.task is not None:
            code.append_crash_context(args.task)
        else:
            code.append_crash_context()
        _append_sys_drgn_task(code)
        code.print()
        return None

    if args.panic:
        prog.config["crash_context"] = _crash_get_panic_context(prog)
        prog.config.pop("crash_context_origin", None)
    elif args.cpu is not None:
        prog.config["crash_context"] = cpu_curr(prog, args.cpu)
        prog.config["crash_context_origin"] = ("cpu", args.cpu)
    elif args.task is not None:
        prog.config["crash_context"] = crash_get_context(prog, args.task)
        prog.config["crash_context_origin"] = args.task
    return _print_sys(prog, system_fields=False, context="current")


@_crash_foreach_subcommand(arguments=(drgn_argument,))
def _crash_foreach_set(task_selector: _TaskSelector, args: argparse.Namespace) -> None:
    prog = task_selector.prog

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            # No append_task_header() because it is a subset of
            # _append_sys_drgn_task() anyways.
            _append_sys_drgn_task(code)
        return code.print()

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()
        print_task_header(task)
        _print_sys(prog, system_fields=False, context=task)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Program Decorators
------------------

The ``drgn.helpers.common.prog`` module provides decorators to transparently
use the :ref:`default program argument <default-program>`.
"""

import functools
import inspect
import typing

from drgn import Object, Program, get_default_prog

__all__ = (
    "takes_object_or_program_or_default",
    "takes_program_or_default",
)

# We don't need any of this at runtime.
if typing.TYPE_CHECKING:
    import sys
    from typing import Any, Optional, Protocol, TypeVar, overload  # novermin

    if sys.version_info < (3, 10):
        from typing_extensions import Callable, Concatenate, ParamSpec
    else:
        from typing import Callable, Concatenate, ParamSpec  # novermin

    P = ParamSpec("P")
    R = TypeVar("R")
    R_co = TypeVar("R_co", covariant=True)

    class TakesProgram(Protocol[P, R_co]):
        def __call__(
            self, prog: Program, *args: P.args, **kwargs: P.kwargs
        ) -> R_co: ...

    class TakesProgramOrDefault(Protocol[P, R_co]):
        @overload
        def __call__(
            self, prog: Program, *args: P.args, **kwargs: P.kwargs
        ) -> R_co: ...

        @overload
        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R_co: ...

    class TakesObjectOrProgramOrDefault(Protocol[P, R_co]):
        @overload
        def __call__(
            self, prog: Program, *args: P.args, **kwargs: P.kwargs
        ) -> R_co: ...

        @overload
        def __call__(
            self, __obj: Object, *args: P.args, **kwargs: P.kwargs
        ) -> R_co: ...

        @overload
        def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R_co: ...


def takes_program_or_default(f: "TakesProgram[P, R]") -> "TakesProgramOrDefault[P, R]":
    """
    Wrap a function taking a :class:`~drgn.Program` so that it uses the
    :ref:`default program argument <default-program>` if omitted.

    .. code-block:: python3

        @takes_program_or_default
        def my_helper(prog: Program, n: IntegerLike) -> Foo:
            ...

        my_helper(1)
        # is equivalent to
        my_helper(get_default_prog(), 1)

        obj = Object(...)
        my_helper(obj)
        # is equivalent to
        my_helper(obj.prog_, obj)
    """
    parameters_iter = iter(inspect.signature(f).parameters.values())
    if next(parameters_iter).name != "prog":
        raise TypeError("first parameter must be prog: Program")
    param1 = None
    for parameter in parameters_iter:
        if parameter.kind not in (
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.VAR_KEYWORD,
        ):
            param1 = parameter.name
        break

    if param1 is None:

        @functools.wraps(f)
        def wrapper(*args: "Any", **kwds: "Any") -> "R":
            if (args and isinstance(args[0], Program)) or ("prog" in kwds):
                return f(*args, **kwds)
            else:
                return f(get_default_prog(), *args, **kwds)

    else:

        @functools.wraps(f)
        def wrapper(*args: "Any", **kwds: "Any") -> "R":
            if args:
                if isinstance(args[0], Program):
                    return f(*args, **kwds)
                elif isinstance(args[0], Object):
                    return f(args[0].prog_, *args, **kwds)
            elif "prog" in kwds:
                return f(**kwds)  # type: ignore
            elif param1 in kwds:
                arg1 = kwds[param1]
                if isinstance(arg1, Object):
                    return f(arg1.prog_, **kwds)  # type: ignore
            return f(get_default_prog(), *args, **kwds)

    # Update the docstring for pydoc.
    if wrapper.__doc__ is not None:
        wrapper.__doc__ += ":param prog: Program, which may be omitted to use the default program argument.\n"
    return wrapper


def takes_object_or_program_or_default(
    # This should be a Protocol instead of Callable, but there's currently no
    # way for a Protocol to express that the second parameter can have any name
    # and then use that name in the return type. See python/typing#1505.
    f: "Callable[Concatenate[Program, Optional[Object], P], R]",
) -> "TakesObjectOrProgramOrDefault[P, R]":
    """
    Wrap a function taking a :class:`~drgn.Program` and an optional
    :class:`~drgn.Object` so that it accepts a ``Program`` *or* an ``Object``
    *or neither*, in which case the :ref:`default program argument
    <default-program>` is used.

    .. code-block:: python3

        @takes_object_or_program_or_default
        def my_helper(prog: Program, obj: Optional[Object], n: IntegerLike) -> Foo:
            ...

        my_helper(prog, 1)
        # is equivalent to
        my_helper.__wrapped__(prog, None, 1)

        obj = Object(...)
        my_helper(obj, 1)
        # is equivalent to
        my_helper.__wrapped__(obj.prog_, obj, 1)

        my_helper(1)
        # is equivalent to
        my_helper.__wrapped__(get_default_prog(), None, 1)

        one_obj = Object(..., 1)
        my_helper(one_obj)
        # is equivalent to
        my_helper.__wrapped__(one_obj.prog_, None, one_obj)

    .. warning::

        This cannot be used with positional parameters with a default value, as
        that would create ambiguity. Keyword-only parameters with a default
        value are OK.

        .. code-block:: python3

            # NOT ALLOWED
            @takes_object_or_program_or_default
            def my_helper(prog: Program, obj: Optional[Object], foo: str = ""): ...

            # OK
            @takes_object_or_program_or_default
            def my_helper(prog: Program, obj: Optional[Object], *, foo: str = ""): ...

    .. note::

        The object parameter can be passed as a keyword, but because of
        `limitations of the Python type system
        <https://github.com/python/typing/issues/1505>`_, type checkers do
        not recognize this.
    """
    signature = inspect.signature(f)
    parameters = list(signature.parameters.values())
    if parameters[0].name != "prog":
        raise TypeError("first parameter must be prog: Program")
    object_param = parameters[1].name
    extra_params = []
    for parameter in parameters[2:]:
        if parameter.kind in (
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.VAR_KEYWORD,
        ):
            break
        if parameter.default != inspect.Parameter.empty:
            raise ValueError(
                f"{getattr(f, '__name__', str(f))} using @takes_object_or_program_or_default can't have positional parameters with defaults due to ambiguity"
            )
        extra_params.append(parameter.name)

    if extra_params:
        param2 = extra_params[0]

        @functools.wraps(f)
        def wrapper(*args: "Any", **kwds: "Any") -> "R":
            if args:
                if len(args) > len(extra_params) or extra_params[len(args) - 1] in kwds:
                    if isinstance(args[0], Program):
                        return f(args[0], None, *args[1:], **kwds)
                    else:
                        return f(args[0].prog_, *args, **kwds)
                elif isinstance(args[0], Object):
                    return f(args[0].prog_, None, *args, **kwds)
            elif object_param in kwds:
                return f(kwds[object_param].prog_, *args, **kwds)
            elif "prog" in kwds:
                kwds[object_param] = None
                return f(*args, **kwds)
            elif param2 in kwds:
                arg2 = kwds[param2]
                if isinstance(arg2, Object):
                    return f(arg2.prog_, None, *args, **kwds)
            return f(get_default_prog(), None, *args, **kwds)

    else:

        @functools.wraps(f)
        def wrapper(*args: "Any", **kwds: "Any") -> "R":
            if args:
                if isinstance(args[0], Program):
                    return f(args[0], None, *args[1:], **kwds)
                else:
                    return f(args[0].prog_, *args, **kwds)
            elif object_param in kwds:
                return f(kwds[object_param].prog_, *args, **kwds)
            elif "prog" in kwds:
                kwds[object_param] = None
                return f(*args, **kwds)
            return f(get_default_prog(), None, *args, **kwds)

    # Update the signature for pydoc.
    wrapper.__signature__ = signature.replace(  # type: ignore[attr-defined]
        parameters=[
            parameters[1].replace(annotation=typing.Union[Object, Program]),
            *parameters[2:],
        ]
    )
    return wrapper



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Get a stack trace from a call to an invalid address on x86-64 or AArch64. drgn
should do this automatically eventually.
"""

import drgn
from drgn import Object
from drgn.helpers.common.prog import takes_program_or_default


def pt_regs_members_from_stack_frame(prog, frame):
    regs = frame.registers()
    if prog.platform.arch == drgn.Architecture.X86_64:
        return {
            "r15": regs.get("r15", 0),
            "r14": regs.get("r14", 0),
            "r13": regs.get("r13", 0),
            "r12": regs.get("r12", 0),
            "bp": regs.get("rbp", 0),
            "bx": regs.get("rbx", 0),
            "r11": regs.get("r11", 0),
            "r10": regs.get("r10", 0),
            "r9": regs.get("r9", 0),
            "r8": regs.get("r8", 0),
            "ax": regs.get("rax", 0),
            "cx": regs.get("rcx", 0),
            "dx": regs.get("rdx", 0),
            "si": regs.get("rsi", 0),
            "di": regs.get("rdi", 0),
            "orig_ax": -1,
            "ip": regs.get("rip", 0),
            "cs": regs.get("cs", 0),
            "flags": regs.get("rflags", 0),
            "sp": regs.get("rsp", 0),
            "ss": regs.get("ss", 0),
        }
    elif prog.platform.arch == drgn.Architecture.AARCH64:
        try:
            pc = frame.pc
        except LookupError:
            pc = 0
        return {
            "regs": [regs.get(f"x{i}", 0) for i in range(31)],
            "sp": regs.get("sp", 0),
            "pc": pc,
            "pstate": regs.get("pstate", 0),
        }
    else:
        raise NotImplementedError()


@takes_program_or_default
def call_fault_stack_trace(prog):
    pt_regs_members = pt_regs_members_from_stack_frame(
        prog, prog.crashed_thread().stack_trace()[0]
    )

    if prog.platform.arch == drgn.Architecture.X86_64:
        pt_regs_members["ip"] = prog.read_word(pt_regs_members["sp"]) - 1
        pt_regs_members["sp"] += 8
    elif prog.platform.arch == drgn.Architecture.AARCH64:
        pt_regs_members["pc"] = pt_regs_members["regs"][30] - 4
    else:
        raise NotImplementedError()

    return prog.stack_trace(Object(prog, "struct pt_regs", pt_regs_members))


if __name__ == "__main__":
    trace = call_fault_stack_trace(prog)
    print(trace)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Resource Limits
---------------

The ``drgn.helpers.linux.resource`` module provides helpers for resource limits
(rlimit).
"""

from typing import Dict, NamedTuple, Optional

from drgn import Object

__all__ = ("task_rlimits",)


def task_rlimits(task: Object) -> Dict[str, "Rlimit"]:
    """
    Get the resource limits on a given task.

    >>> task_rlimits(find_task(os.getpid()))
    {'CPU': Rlimit(cur=None, max=None), 'FSIZE': Rlimit(cur=None, max=None), ...}
    >>> task_rlimits(find_task(os.getpid()))["STACK"].cur
    8388608

    :return: Dictionary mapping resource name (`"NOFILE"`, `"CORE"`, etc.) to
        :class:`Rlimit`.
    """
    # XXX: alpha, mips, and sparc have different values for RLIM_INFINITY and
    # RLIMIT_{NOFILE,AS,RSS,NPROC,MEMLOCK}. Once we recognize those
    # architectures, we will have to account for that.
    RLIM_INFINITY = Object(task.prog_, "unsigned long", -1)
    return {
        name: Rlimit(
            None if rlim.rlim_cur == RLIM_INFINITY else rlim.rlim_cur.value_(),
            None if rlim.rlim_max == RLIM_INFINITY else rlim.rlim_max.value_(),
        )
        for name, rlim in zip(
            (
                "CPU",
                "FSIZE",
                "DATA",
                "STACK",
                "CORE",
                "RSS",
                "NPROC",
                "NOFILE",
                "MEMLOCK",
                "AS",
                "LOCKS",
                "SIGPENDING",
                "MSGQUEUE",
                "NICE",
                "RTPRIO",
                "RTTIME",
            ),
            task.signal.rlim.read_(),
        )
    }


class Rlimit(NamedTuple):
    """Resource limit."""

    cur: Optional[int]
    """Soft limit, or ``None`` if unlimited."""

    max: Optional[int]
    """Hard limit, or ``None`` if unlimited."""



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Formatting
----------

The ``drgn.helpers.common.format`` module provides generic helpers for
formatting different things as text.
"""

import collections
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Hashable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    SupportsFloat,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from drgn import IntegerLike, Type

if TYPE_CHECKING:
    from _typeshed import SupportsWrite

__all__ = (
    "CellFormat",
    "RowOptions",
    "decode_enum_type_flags",
    "decode_flags",
    "double_quote_ascii_string",
    "escape_ascii_character",
    "escape_ascii_string",
    "number_in_binary_units",
    "print_table",
)

_T_co = TypeVar("_T_co", covariant=True)


def escape_ascii_character(
    c: int,
    escape_single_quote: bool = False,
    escape_double_quote: bool = False,
    escape_backslash: bool = False,
) -> str:
    """
    Format an ASCII byte value as a character, possibly escaping it.
    Non-printable characters are always escaped. Non-printable characters other
    than ``\\0``, ``\\a``, ``\\b``, ``\\t``, ``\\n``, ``\\v``, ``\\f``, and
    ``\\r`` are escaped in hexadecimal format (e.g., ``\\x7f``). By default,
    printable characters are never escaped.

    :param c: Character to escape.
    :param escape_single_quote: Whether to escape single quotes to ``\\'``.
    :param escape_double_quote: Whether to escape double quotes to ``\\"``.
    :param escape_backslash: Whether to escape backslashes to ``\\\\``.
    """
    if c == 0:
        return r"\0"
    elif c == 7:
        return r"\a"
    elif c == 8:
        return r"\b"
    elif c == 9:
        return r"\t"
    elif c == 10:
        return r"\n"
    elif c == 11:
        return r"\v"
    elif c == 12:
        return r"\f"
    elif c == 13:
        return r"\r"
    elif escape_double_quote and c == 34:
        return r"\""
    elif escape_single_quote and c == 39:
        return r"\'"
    elif escape_backslash and c == 92:
        return r"\\"
    elif 32 <= c <= 126:
        return chr(c)
    else:
        return f"\\x{c:02x}"


def escape_ascii_string(
    buffer: Iterable[int],
    escape_single_quote: bool = False,
    escape_double_quote: bool = False,
    escape_backslash: bool = False,
) -> str:
    """
    Escape an iterable of ASCII byte values (e.g., :class:`bytes` or
    :class:`bytearray`). See :func:`escape_ascii_character()`.

    :param buffer: Byte array to escape.
    """
    return "".join(
        escape_ascii_character(
            c,
            escape_single_quote=escape_single_quote,
            escape_double_quote=escape_double_quote,
            escape_backslash=escape_backslash,
        )
        for c in buffer
    )


def double_quote_ascii_string(buffer: Iterable[int]) -> str:
    """
    Get an iterable of ASCII byte values (e.g., :class:`bytes` or
    :class:`bytearray`) as a double-quoted string.

    This is equivalent to:

    .. code-block:: python3

        '"' + escape_ascii_string(buffer, escape_double_quote=True, escape_backslash=True) + '"'
    """
    parts = [
        escape_ascii_character(c, escape_double_quote=True, escape_backslash=True)
        for c in buffer
    ]
    parts.insert(0, '"')
    parts.append('"')
    return "".join(parts)


def decode_flags(
    value: IntegerLike,
    flags: Iterable[Tuple[str, int]],
    bit_numbers: bool = True,
    aliases: bool = True,
) -> str:
    """
    Get a human-readable representation of a bitmask of flags.

    By default, flags are specified by their bit number:

    >>> decode_flags(2, [("BOLD", 0), ("ITALIC", 1), ("UNDERLINE", 2)])
    'ITALIC'

    They can also be specified by their value:

    >>> decode_flags(2, [("BOLD", 1), ("ITALIC", 2), ("UNDERLINE", 4)],
    ...              bit_numbers=False)
    'ITALIC'

    Multiple flags are combined with "|":

    >>> decode_flags(5, [("BOLD", 0), ("ITALIC", 1), ("UNDERLINE", 2)])
    'BOLD|UNDERLINE'

    If there are multiple names for the same bit, they are all included:

    >>> decode_flags(2, [("SMALL", 0), ("BIG", 1), ("LARGE", 1)])
    'BIG|LARGE'

    If there are any unknown bits, their raw value is included:

    >>> decode_flags(27, [("BOLD", 0), ("ITALIC", 1), ("UNDERLINE", 2)])
    'BOLD|ITALIC|0x18'

    Zero is returned verbatim:

    >>> decode_flags(0, [("BOLD", 0), ("ITALIC", 1), ("UNDERLINE", 2)])
    '0'

    :param value: Bitmask to decode.
    :param flags: List of flag names and their bit numbers or values.
    :param bit_numbers: Whether *flags* specifies the bit numbers (where 0 is
        the least significant bit) or values of the flags.
    :param aliases: Whether to include multiple names for the same flag. If
        ``False``, then only the first name is included.
    """
    value = value.__index__()
    if value == 0:
        return "0"

    parts = []
    orig_value = value
    for name, flag in flags:
        if bit_numbers:
            flag = 1 << flag
        if (orig_value if aliases else value) & flag:
            parts.append(name)
            value &= ~flag

    if value:
        parts.append(hex(value))

    return "|".join(parts)


def decode_enum_type_flags(
    value: IntegerLike,
    type: Type,
    bit_numbers: bool = True,
) -> str:
    """
    Get a human-readable representation of a bitmask of flags where the flags
    are specified by an enumerated :class:`drgn.Type`.

    This supports enums where the values are bit numbers:

    >>> print(bits_enum)
    enum style_bits {
            BOLD = 0,
            ITALIC = 1,
            UNDERLINE = 2,
    }
    >>> decode_enum_type_flags(5, bits_enum)
    'BOLD|UNDERLINE'

    Or the values of the flags:

    >>> print(flags_enum)
    enum style_flags {
            BOLD = 1,
            ITALIC = 2,
            UNDERLINE = 4,
    }
    >>> decode_enum_type_flags(5, flags_enum, bit_numbers=False)
    'BOLD|UNDERLINE'

    See :func:`decode_flags()`.

    :param value: Bitmask to decode.
    :param type: Enumerated type with bit numbers for enumerators.
    :param bit_numbers: Whether the enumerator values specify the bit numbers
         or values of the flags.
    """
    enumerators = type.enumerators
    if enumerators is None:
        raise TypeError("cannot decode incomplete enumerated type")
    return decode_flags(
        value,
        enumerators,  # type: ignore  # python/mypy#592
        bit_numbers,
    )


def number_in_binary_units(n: SupportsFloat, precision: int = 1) -> str:
    """
    Format a number in binary units (i.e., "K" is 1024, "M" is 1024\\ :sup:`2`,
    etc.).

    >>> number_in_binary_units(1280)
    '1.2K'

    A precision can be specified:

    >>> number_in_binary_units(1280, precision=2)
    '1.25K'

    Exact numbers are printed without a fractional part:

    >>> number_in_binary_units(1024 * 1024)
    '1M'

    Numbers less than 1024 are not scaled:

    >>> number_in_binary_units(10)
    '10'

    :param n: Number to format.
    :param precision: Number of digits to include in fractional part.
    """
    n = float(n)
    for prefix in ("", "K", "M", "G", "T", "P", "E", "Z"):
        if abs(n) < 1024:
            break
        n /= 1024.0
    else:
        prefix = "Y"
    if n.is_integer():
        precision = 0
    return f"{n:.{precision}f}{prefix}"


def _print_table_row(
    row: Sequence[Any],
    *,
    widths: Sequence[int],
    sep: str = "  ",
    file: "Optional[SupportsWrite[str]]" = None,
) -> None:
    print(
        *(
            # `or ''` works around
            # https://github.com/python/cpython/issues/71959 on Python < 3.10.
            f"{value:{widths[i] or ''}}".rstrip(" " if i == len(row) - 1 else "")
            for i, value in enumerate(row)
        ),
        sep=sep,
        file=file,
    )


def print_table(
    rows: Sequence[Sequence[Any]],
    *,
    sep: str = "  ",
    file: "Optional[SupportsWrite[str]]" = None,
) -> None:
    """
    Print data as a table.

    The input is given as a sequence (e.g., :class:`list` or :class:`tuple`) of
    rows, where each row is a sequence of values. Rows can have different
    lengths.

    >>> print_table([[2, 2000, 4], ["", 3, 13, 19]])
    2  2000   4
          3  13  19

    By default, numbers are right-aligned and most other objects are
    left-aligned. This (and other format options) can be changed by wrapping
    the value in a :class:`CellFormat`.

    >>> print_table(
    ...     [
    ...         ["DECIMAL", "HEXADECIMAL"],
    ...         [CellFormat(10, "<"), CellFormat(10, "<x")],
    ...     ]
    ... )
    DECIMAL  HEXADECIMAL
    10       a

    Rows can also be aligned in separate groups by wrapping them in a
    :class:`RowOptions`. The default group is 0.

    >>> print_table(
    ...     [
    ...         ["LETTER", "INDEX"],
    ...         ["A", 1],
    ...         RowOptions(["WORD", "LENGTH"], group=1),
    ...         RowOptions(["angstrom", 8], group=1),
    ...         ["LETTER", "INDEX"],
    ...         ["B", 2],
    ...         RowOptions(["WORD", "LENGTH"], group=1),
    ...         RowOptions(["banana", 6], group=1),
    ...     ]
    ... )
    ...
    LETTER  INDEX
    A           1
    WORD      LENGTH
    angstrom       8
    LETTER  INDEX
    B           2
    WORD      LENGTH
    banana         6

    :param rows: Sequence of rows, where each row is a sequence of cells.
    :param sep: Column separator.
    :param file: File to write to (defaults to ``sys.stdout``).
    """
    group_widths: Dict[Any, List[int]] = collections.defaultdict(list)
    for row in rows:
        widths = group_widths[row._group if isinstance(row, RowOptions) else 0]
        for i, value in enumerate(row):
            cell_width = len(str(value))
            if i < len(widths):
                widths[i] = max(widths[i], cell_width)
            else:
                widths.append(cell_width)

    for row in rows:
        _print_table_row(
            row,
            widths=group_widths[row._group if isinstance(row, RowOptions) else 0],
            sep=sep,
            file=file,
        )


class CellFormat:
    _FORMAT_SPEC_RE = re.compile(
        r"""
        (?P<options>
            (?:
                .?      # fill
                [<>=^]  # align
            )?
            [-+ ]?  # sign
            z?
            [#]?
            0?
        )
        (?P<width>[0-9]+)?
        (?P<rest>
            [,_]?               # grouping
            (?:\.[0-9])?        # precision
            [bcdeEfFgGnosxX%]?  # type
        )
        """,
        flags=re.VERBOSE,
    )

    def __init__(self, value: Any, format_spec: str) -> None:
        """
        Wrap a value with additional format options to apply when it is
        formatted by :func:`print_table()`.

        :param value: Value to wrap.
        :param format_spec: :ref:`Format specification <formatspec>`. It may
            not specify a width.
        """
        self._value = value
        match = self._FORMAT_SPEC_RE.fullmatch(format_spec)
        if not match:
            raise ValueError(f"invalid format_spec {format_spec!r}")
        if match.group("width"):
            raise ValueError("format_spec must not have width")
        self._options = match.group("options")
        self._rest = match.group("rest")

    def __str__(self) -> str:
        return f"{self._value:{self._options}{self._rest}}"

    def __format__(self, format_spec: str) -> str:
        return f"{self._value:{self._options}{format_spec}{self._rest}}"


class RowOptions(Sequence[_T_co]):
    def __init__(self, row: Sequence[_T_co], *, group: Hashable = 0) -> None:
        """
        Wrap a row with additional options to apply when it is formatted by
        :func:`print_table()`.

        :param row: Row to wrap.
        :param group: Align this row only with other rows in this group.
        """
        self._seq = row
        self._group = group

    def __iter__(self) -> Iterator[_T_co]:
        return iter(self._seq)

    def __contains__(self, x: object, /) -> bool:
        return x in self._seq

    def __len__(self) -> int:
        return len(self._seq)

    def __reversed__(self) -> Iterator[_T_co]:
        return reversed(self._seq)

    @overload
    def __getitem__(self, index: int) -> _T_co: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[_T_co]: ...

    def __getitem__(self, index: Union[int, slice]) -> Union[_T_co, Sequence[_T_co]]:
        return self._seq[index]

    def index(self, *args: Any) -> int:
        return self._seq.index(*args)

    def count(self, value: Any) -> int:
        return self._seq.count(value)



#!/usr/bin/env drgn
# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Dump Btrfs subvolumes that have been deleted but not cleaned up."""

from drgn import Object, cast
from drgn.helpers.linux.fs import inode_path, path_lookup
from drgn.helpers.linux.radixtree import radix_tree_for_each
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry


def dump_orphan_subvolumes(fs_info: Object) -> None:
    prog = fs_info.prog_
    BTRFS_ROOT_ORPHAN_ITEM_INSERTED = prog["BTRFS_ROOT_ORPHAN_ITEM_INSERTED"]
    for objectid, entry in radix_tree_for_each(fs_info.fs_roots_radix):
        root = cast("struct btrfs_root *", entry)
        if root.state & (1 << BTRFS_ROOT_ORPHAN_ITEM_INSERTED):
            print(f"orphan root {objectid} has the following inodes in memory:")
            for inode in rbtree_inorder_for_each_entry(
                "struct btrfs_inode", root.inode_tree.address_of_(), "rb_node"
            ):
                path = inode_path(inode.vfs_inode.address_of_())
                if path is None:
                    print(f"    inode {inode.vfs_inode.i_ino.value_()} with no cached names")
                else:
                    print(f"    {path.decode()}")


if __name__ == "__main__":
    import argparse
    from pathlib import Path

    parser = argparse.ArgumentParser()
    parser.add_argument("path", type=Path)
    args = parser.parse_args()

    dump_orphan_subvolumes(
        cast(
            "struct btrfs_fs_info *",
            path_lookup(prog, args.path.resolve()).mnt.mnt_sb.s_fs_info,
        )
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Sbitmaps
--------

The ``drgn.helpers.linux.sbitmap`` module provides helpers for working with
scalable bitmaps from :linux:`include/linux/sbitmap.h`.
"""

from typing import Iterator

from drgn import Object

__all__ = ("sbitmap_for_each_set",)


def sbitmap_for_each_set(sb: Object) -> Iterator[int]:
    """
    Iterate over all set (one) bits in a sbitmap.

    :param sb: ``struct sbitmap *``
    """
    depth = sb.depth.value_()
    shift = sb.shift.value_()
    map_nr = sb.map_nr.value_()
    for i, map in enumerate(sb.map[:map_nr]):
        map_start = i << shift
        if i < map_nr - 1:
            map_depth = 1 << shift
        else:
            map_depth = depth - map_start

        word = map.word.value_()
        # sbitmap_word::cleared was added in Linux kernel commit ("sbitmap:
        # ammortize cost of clearing bits") (in v5.0).
        try:
            word &= ~map.cleared.value_()
        except AttributeError:
            pass
        for j in range(map_depth):
            if word & (1 << j):
                yield map_start + j



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Timers
------

The ``drgn.helpers.linux.timer`` module provides helpers for kernel timers,
including the timer wheel and high-resolution timers ("hrtimers").
"""

from typing import Iterator, Sequence

from drgn import Object, Program
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.list import hlist_for_each_entry
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry

__all__ = (
    "hrtimer_clock_base_for_each",
    "timer_base_for_each",
    "timer_base_names",
)


@takes_program_or_default
def timer_base_names(prog: Program) -> Sequence[str]:
    """
    Get the names of the timer wheel bases used on this kernel.

    This depends on the kernel version and configuration. For example:

    >>> timer_base_names()
    ('BASE_LOCAL', 'BASE_GLOBAL', 'BASE_DEF')

    means that each CPU has 3 timer bases: ``BASE_LOCAL`` (0), ``BASE_GLOBAL``
    (1), and ``BASE_DEF`` (2).
    """
    try:
        return prog.cache["timer_base_names"]
    except KeyError:
        pass
    nr_bases = len(prog["timer_bases"])
    if nr_bases == 3:
        # Since Linux kernel commit 83a665dc99a7 ("timers: Keep the pinned
        # timers separate from the others") (in v6.9) with
        # CONFIG_NO_HZ_COMMON=y.
        base_names: Sequence[str] = ("BASE_LOCAL", "BASE_GLOBAL", "BASE_DEF")
    elif nr_bases == 2:
        # Before that commit with CONFIG_NO_HZ_COMMON=y.
        base_names = ("BASE_STD", "BASE_DEF")
    elif nr_bases == 1:
        # CONFIG_NO_HZ_COMMON=n. We can't directly detect the above commit in
        # this case, so we have to differentiate based on the next commit,
        # 21927fc89e5f ("timers: Retrieve next expiry of pinned/non-pinned
        # timers separately"), which added the struct timer_events type.
        try:
            prog.type("struct timer_events")
            base_names = ("BASE_LOCAL",)
        except LookupError:
            base_names = ("BASE_STD",)
    else:
        raise NotImplementedError(f"unknown NR_BASES {nr_bases}")
    prog.cache["timer_base_names"] = base_names
    return base_names


def timer_base_for_each(base: Object) -> Iterator[Object]:
    """
    Iterate over every timer on a timer base.

    .. code-block:: python3

        for cpu in for_each_online_cpu():
            for base in per_cpu(prog["timer_bases"], cpu):
                for timer in timer_base_for_each(base.address_of_()):
                    ...

    :param base: ``struct timer_base *``
    :return: Iterator of ``struct timer_list *`` objects.
    """
    for head in base.vectors:
        yield from hlist_for_each_entry(
            "struct timer_list", head.address_of_(), "entry"
        )


def hrtimer_clock_base_for_each(clock_base: Object) -> Iterator[Object]:
    """
    Iterate over every high-resolution timer on an hrtimer clock base.

    .. code-block:: python3

        for cpu in for_each_online_cpu():
            for cpu_base in per_cpu(prog["hrtimer_bases"], cpu):
                for clock_base in cpu_base.clock_base:
                    for hrtimer in hrtimer_clock_base_for_each(clock_base.address_of_()):
                        ...

    :param clock_base: ``struct hrtimer_clock_base *``
    :return: Iterator of ``struct hrtimer *`` objects.
    """
    return rbtree_inorder_for_each_entry(
        "struct hrtimer", clock_base.active.rb_root.rb_root.address_of_(), "node.node"
    )
